Imports System.Reflection
Imports System.ComponentModel
Imports System.IO
Imports System.Runtime.InteropServices
Imports System.Drawing.Drawing2D
Imports System.Windows.Forms.VisualStyles

Imports Stax.Win32
Imports Stax.UI.Design
Imports Stax.AppFramework

Namespace UI
    Public Class ExtendedTreeView
        Inherits TreeView

        Public Event CustomPaint(ByVal e As PaintEventArgs)

        Private AutoCollapsValue As Boolean

        <DefaultValue(False)> _
        Public Property AutoCollaps() As Boolean
            Get
                Return AutoCollapsValue
            End Get
            Set(ByVal Value As Boolean)
                AutoCollapsValue = Value
            End Set
        End Property

        Private ExpandModeValue As TreeNodeExpandMode

        <DefaultValue(GetType(TreeNodeExpandMode), "Disabled")> _
        Public Property ExpandMode() As TreeNodeExpandMode
            Get
                Return ExpandModeValue
            End Get
            Set(ByVal Value As TreeNodeExpandMode)
                ExpandModeValue = Value
            End Set
        End Property

        Private SelectOnMouseDownValue As Boolean

        <DefaultValue(False)> _
        Public Property SelectOnMouseDown() As Boolean
            Get
                Return SelectOnMouseDownValue
            End Get
            Set(ByVal Value As Boolean)
                SelectOnMouseDownValue = Value
            End Set
        End Property

        Protected Overrides Sub OnAfterSelect(ByVal e As TreeViewEventArgs)
            If AutoCollaps Then
                Dim count As Integer

                For Each i As TreeNode In Me.Nodes
                    If i.IsExpanded Then
                        count += 1
                    End If
                Next

                If count < 2 Then
                    Dim rootNode As TreeNode = e.Node

                    While True
                        If rootNode.Parent Is Nothing Then
                            Exit While
                        Else
                            rootNode = rootNode.Parent
                        End If
                    End While

                    For Each i As TreeNode In Nodes
                        If Not i Is rootNode Then
                            i.Collapse()
                        End If
                    Next
                End If
            End If

            Select Case ExpandMode
                Case TreeNodeExpandMode.Normal
                    e.Node.Expand()
                Case TreeNodeExpandMode.InclusiveChilds
                    e.Node.ExpandAll()
            End Select

            MyBase.OnAfterSelect(e)
        End Sub

        Protected Overrides Sub OnMouseDown(ByVal e As MouseEventArgs)
            MyBase.OnMouseDown(e)

            If SelectOnMouseDown Then
                Dim n = GetNodeAt(ClientMousePos)

                If Not n Is Nothing AndAlso n.Nodes.Count = 0 Then
                    SelectedNode = n
                End If
            End If
        End Sub

        Public Sub MoveSelectionLeft()
            Dim n As TreeNode = SelectedNode

            If Not n Is Nothing AndAlso Not n.Parent Is Nothing Then
                Dim parentParentNodes As TreeNodeCollection = GetParentParentNodes(n)
                Dim parentIndex As Integer = n.Parent.Index

                If Not parentParentNodes Is Nothing Then
                    n.Remove()
                    parentParentNodes.Insert(parentIndex + 1, n)
                    SelectedNode = n
                End If
            End If
        End Sub

        Public Sub MoveSelectionUp()
            Dim n As TreeNode = SelectedNode

            If Not n Is Nothing Then
                If n.Index = 0 Then
                    Dim parentParentNodes As TreeNodeCollection = GetParentParentNodes(n)

                    If Not parentParentNodes Is Nothing Then
                        Dim index As Integer = n.Parent.Index
                        n.Remove()
                        parentParentNodes.Insert(index, n)
                    End If
                Else
                    Dim index As Integer = n.Index
                    Dim parentNodes As TreeNodeCollection = GetParentNodes(n)
                    n.Remove()
                    parentNodes.Insert(index - 1, n)
                End If

                SelectedNode = n
            End If
        End Sub

        Public Sub MoveSelectionRight()
            Dim n As TreeNode = SelectedNode

            If Not n Is Nothing Then
                If n.Index > 0 Then
                    Dim previousNode As TreeNode = n.PrevNode
                    n.Remove()
                    previousNode.Nodes.Add(n)
                    SelectedNode = n
                End If
            End If
        End Sub

        Public Sub MoveSelectionDown()
            Dim n As TreeNode = SelectedNode

            If Not n Is Nothing Then
                If n.NextNode Is Nothing Then
                    Dim parentParentNodes As TreeNodeCollection = GetParentParentNodes(n)

                    If Not parentParentNodes Is Nothing Then
                        Dim index As Integer = n.Parent.Index
                        n.Remove()
                        parentParentNodes.Insert(index + 1, n)
                    End If
                Else
                    Dim index As Integer = n.Index
                    Dim parentNodes As TreeNodeCollection = GetParentNodes(n)
                    n.Remove()
                    parentNodes.Insert(index + 1, n)
                End If

                SelectedNode = n
            End If
        End Sub

        Private Function GetParentParentNodes(ByVal n As TreeNode) As TreeNodeCollection
            Dim parent As TreeNode = n.Parent

            If parent Is Nothing Then
                Return Nothing
            End If

            If parent.Parent Is Nothing Then
                Return Nodes
            Else
                Return parent.Parent.Nodes
            End If
        End Function

        Private Function GetParentNodes(ByVal n As TreeNode) As TreeNodeCollection
            Dim parent As TreeNode = n.Parent

            If parent Is Nothing Then
                Return Nodes
            Else
                Return parent.Nodes
            End If
        End Function
    End Class

    Public Enum TreeNodeExpandMode
        Disabled
        Normal
        InclusiveChilds
    End Enum

    Public Class ExtendedToolStrip
        Inherits ToolStrip

        Private IsAppThemed As Boolean = VisualStyleInformation.IsEnabledByUser AndAlso Application.RenderWithVisualStyles

        Public Sub New()
            MyBase.New()
        End Sub

        Public Sub New(ByVal ParamArray items As ToolStripItem())
            MyBase.New(items)
        End Sub

        Private OverdrawLineValue As Boolean

        <DefaultValue(False), Description("Overdraws a weird grey line visible at the bottom in some configurations.")> _
        Public Property OverdrawLine() As Boolean
            Get
                Return OverdrawLineValue
            End Get
            Set(ByVal Value As Boolean)
                OverdrawLineValue = Value
                Invalidate()
            End Set
        End Property

        Private SingleAutoCheckedValue As Boolean

        <DefaultValue(False), Description("Only one button can be checked at the same time.")> _
        Public Property SingleAutoChecked() As Boolean
            Get
                Return SingleAutoCheckedValue
            End Get
            Set(ByVal Value As Boolean)
                SingleAutoCheckedValue = Value
            End Set
        End Property

        Private ShowControlBorderValue As Boolean

        <DefaultValue(False), Description("Shows a themed control border.")> _
        Public Property ShowControlBorder() As Boolean
            Get
                Return ShowControlBorderValue
            End Get
            Set(ByVal Value As Boolean)
                ShowControlBorderValue = Value
            End Set
        End Property

        Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
            MyBase.OnPaint(e)

            If OverdrawLine Then
                Dim b As New SolidBrush(BackColor)
                e.Graphics.FillRectangle(b, 0, Height - 2, Width, 2)
                b.Dispose()
            End If

            If ShowControlBorder AndAlso IsAppThemed Then
                ControlPaint.DrawBorder(e.Graphics, ClientRectangle, VisualStyleInformation.TextControlBorder, ButtonBorderStyle.Solid)
            End If
        End Sub

        Protected Overrides Sub OnItemClicked(ByVal e As ToolStripItemClickedEventArgs)
            If SingleAutoChecked Then
                For Each i As ToolStripItem In Items
                    If TypeOf i Is ToolStripButton Then
                        If i Is e.ClickedItem Then
                            CType(i, ToolStripButton).Checked = True
                        Else
                            CType(i, ToolStripButton).Checked = False
                        End If
                    End If
                Next
            End If

            MyBase.OnItemClicked(e)
        End Sub

        Protected Overrides ReadOnly Property CreateParams() As CreateParams
            Get
                Dim ret = MyBase.CreateParams

                If ShowControlBorder AndAlso Not IsAppThemed Then
                    ret.ExStyle = ret.ExStyle Or CInt(WS.EX_CLIENTEDGE)
                End If

                Return ret
            End Get
        End Property
    End Class

    Public Class ProxyControlContainer
        Inherits Panel

        Private ActiveControl As Type
        Private ChildControls As New Dictionary(Of Type, Control)

        Public Function ShowControl(ByVal t As Type, ByVal show As Boolean) As Control
            If Not ActiveControl Is t Then
                If Not ChildControls.ContainsKey(t) Then
                    Dim c = DirectCast(Activator.CreateInstance(t), Control)
                    c.Visible = show
                    ChildControls(t) = c
                    c.Dock = DockStyle.Fill
                    Controls.Add(c)
                End If

                If show Then
                    If Not ChildControls(t).Visible Then
                        ChildControls(t).Visible = True
                    End If

                    ChildControls(t).BringToFront()
                    ActiveControl = t
                End If
            End If

            Return ChildControls(t)
        End Function
    End Class

    <ToolboxItem(False)> _
    Public Class DialogControl
        Inherits UserControl

        Private IsAppThemed As Boolean = VisualStyleInformation.IsEnabledByUser AndAlso Application.RenderWithVisualStyles

        Public Event FormClosingOK()

        Public Sub New()
            MyBase.New()
            Padding = New Padding(8)
            SetStyle(ControlStyles.CacheText, True)
            BackColor = Color.Transparent
        End Sub

        <Category("Appearance"), DefaultValue(""), Bindable(True), _
        DesignerSerializationVisibility(DesignerSerializationVisibility.Visible), _
        Browsable(True), EditorBrowsable(EditorBrowsableState.Always)> _
        Public Overrides Property Text() As String
            Get
                Return MyBase.Text
            End Get
            Set(ByVal Value As String)
                If Value <> MyBase.Text Then
                    MyBase.Text = Value
                    Dim p As Padding = Padding

                    If Value = "" Then
                        p.Top = 3
                    Else
                        p.Top = Font.Height + 15
                    End If

                    If Not p = Padding Then
                        Padding = p
                    End If

                    Invalidate()
                End If
            End Set
        End Property

        Private BackgroundRendererValue As VisualStyleRenderer

        Private ReadOnly Property BackgroundRenderer() As VisualStyleRenderer
            Get
                If BackgroundRendererValue Is Nothing AndAlso IsAppThemed Then
                    Dim e = VisualStyleElement.Tab.Pane.Normal
                    BackgroundRendererValue = New VisualStyleRenderer(e.ClassName, e.Part, 0)
                End If

                Return BackgroundRendererValue
            End Get
        End Property

        Protected Overrides Sub OnPaintBackground(ByVal e As PaintEventArgs)
            If Not BackgroundRenderer Is Nothing Then
                BackgroundRenderer.DrawBackground(e.Graphics, ClientRectangle)
            Else
                MyBase.OnPaintBackground(e)
            End If

            If Text <> "" Then
                Using b As New SolidBrush(ForeColor)
                    Dim f = New Font(Font, FontStyle.Bold)
                    e.Graphics.DrawString(Text, f, b, Padding.Left + 3, 6)
                End Using

                Dim size As SizeF = e.Graphics.MeasureString(Text, Font)

                If Not BackgroundRenderer Is Nothing Then
                    Dim rect As New Rectangle(Padding.Left + 3, Font.Height + 10, Width - Padding.Left - Padding.Right - 6, 1)
                    Dim p As New LinearGradientBrush(rect, SystemColors.ControlDarkDark, SystemColors.ControlLight, LinearGradientMode.Horizontal)
                    e.Graphics.FillRectangle(p, rect)
                    p.Dispose()
                Else
                    If Enabled Then
                        Dim p As New Pen(SystemColors.ControlDark)
                        e.Graphics.DrawLine(p, Padding.Left + 3, Font.Height + 10, Width - Padding.Right - 3, Font.Height + 10)
                        p.Dispose()

                        e.Graphics.DrawLine(Pens.White, Padding.Left + 3, Font.Height + 11, Width - Padding.Right - 3, Font.Height + 11)

                        Dim offset As Integer = 1
                        Dim visibleWidth As Integer = Width - 5
                    Else
                        Dim p As New Pen(SystemColors.InactiveBorder)
                        e.Graphics.DrawLine(p, Padding.Left + 3, Font.Height + 11, Width - Padding.Right - 3, Font.Height + 11)
                        p.Dispose()
                    End If
                End If
            End If
        End Sub

        Protected Overrides Sub OnHandleCreated(ByVal e As EventArgs)
            MyBase.OnHandleCreated(e)

            If Not DesignMode Then
                AddHandler FindForm.FormClosing, AddressOf FormClosing
            End If
        End Sub

        Private Sub FormClosing(ByVal sender As Object, ByVal e As FormClosingEventArgs)
            If DirectCast(sender, Form).DialogResult = DialogResult.OK Then
                OnFormClosingOK()
            End If
        End Sub

        Protected Overridable Sub OnFormClosingOK()
            RaiseEvent FormClosingOK()
        End Sub

        Public Overridable Sub ShowHelp()
        End Sub
    End Class

    <ToolboxItem(False)> _
    Public Class DialogPanel
        Inherits Panel

        Private IsAppThemed As Boolean = VisualStyleInformation.IsEnabledByUser AndAlso Application.RenderWithVisualStyles

        Public Event FormClosingOK()

        Public Sub New()
            MyBase.New()
            Padding = New Padding(8)
            SetStyle(ControlStyles.CacheText, True)
        End Sub

        <Category("Appearance"), DefaultValue(""), Bindable(True), _
        DesignerSerializationVisibility(DesignerSerializationVisibility.Visible), _
        Browsable(True), EditorBrowsable(EditorBrowsableState.Always)> _
        Public Overrides Property Text() As String
            Get
                Return MyBase.Text
            End Get
            Set(ByVal Value As String)
                If Value <> MyBase.Text Then
                    MyBase.Text = Value
                    Dim p As Padding = Padding

                    If Value = "" Then
                        p.Top = 3
                    Else
                        p.Top = Font.Height + 15
                    End If

                    If Not p = Padding Then
                        Padding = p
                    End If

                    Invalidate()
                End If
            End Set
        End Property

        Private TabPageStyleValue As Boolean

        <DefaultValue(False)> _
        Public Property UseVisualStyle() As Boolean
            Get
                Return TabPageStyleValue
            End Get
            Set(ByVal Value As Boolean)
                If Value <> TabPageStyleValue Then
                    TabPageStyleValue = Value

                    If Value Then
                        BackColor = Color.Transparent
                    End If

                    Invalidate()
                End If
            End Set
        End Property

        Private BackgroundRendererValue As VisualStyleRenderer

        Private ReadOnly Property BackgroundRenderer() As VisualStyleRenderer
            Get
                If BackgroundRendererValue Is Nothing AndAlso IsAppThemed AndAlso UseVisualStyle Then
                    Dim e As VisualStyleElement = VisualStyleElement.Tab.Pane.Normal
                    BackgroundRendererValue = New VisualStyleRenderer(e.ClassName, e.Part, 0)
                End If

                Return BackgroundRendererValue
            End Get
        End Property

        Protected Overrides Sub OnPaintBackground(ByVal e As PaintEventArgs)
            If Not BackgroundRenderer Is Nothing Then
                BackgroundRenderer.DrawBackground(e.Graphics, ClientRectangle)
            Else
                MyBase.OnPaintBackground(e)
            End If

            If Text <> "" Then
                Dim b As New SolidBrush(ForeColor)
                Dim f As Font = New Font(Font, FontStyle.Bold)
                e.Graphics.DrawString(Text, f, b, Padding.Left + 3, 6)
                b.Dispose()

                Dim size As SizeF = e.Graphics.MeasureString(Text, Font)

                If Not BackgroundRenderer Is Nothing Then
                    Dim rect As New Rectangle(Padding.Left + 3, Font.Height + 10, Width - Padding.Left - Padding.Right - 6, 1)
                    Dim p As New LinearGradientBrush(rect, SystemColors.ControlDarkDark, SystemColors.ControlLight, LinearGradientMode.Horizontal)
                    e.Graphics.FillRectangle(p, rect)
                    p.Dispose()
                Else
                    If Enabled Then
                        Dim p As New Pen(SystemColors.ControlDark)
                        e.Graphics.DrawLine(p, Padding.Left + 3, Font.Height + 10, Width - Padding.Right - 3, Font.Height + 10)
                        p.Dispose()

                        e.Graphics.DrawLine(Pens.White, Padding.Left + 3, Font.Height + 11, Width - Padding.Right - 3, Font.Height + 11)

                        Dim offset As Integer = 1
                        Dim visibleWidth As Integer = Width - 5
                    Else
                        Dim p As New Pen(SystemColors.InactiveBorder)
                        e.Graphics.DrawLine(p, Padding.Left + 3, Font.Height + 11, Width - Padding.Right - 3, Font.Height + 11)
                        p.Dispose()
                    End If
                End If
            End If
        End Sub

        Protected Overrides Sub OnHandleCreated(ByVal e As EventArgs)
            MyBase.OnHandleCreated(e)

            If Not DesignMode Then
                AddHandler FindForm.FormClosing, AddressOf FormClosing
            End If
        End Sub

        Private Sub FormClosing(ByVal sender As Object, ByVal e As FormClosingEventArgs)
            If DirectCast(sender, Form).DialogResult = DialogResult.OK Then
                OnFormClosingOK()
            End If
        End Sub

        Protected Overridable Sub OnFormClosingOK()
            RaiseEvent FormClosingOK()
        End Sub
    End Class

    <DesignerAttribute(GetType(LineControlDesigner))> _
    Public Class LineControl
        Inherits UserControl

        Private IsGradientValue As Boolean

        <DefaultValue(False)> _
        Public Property IsGradient() As Boolean
            Get
                Return IsGradientValue
            End Get
            Set(ByVal Value As Boolean)
                IsGradientValue = Value
                Invalidate()
            End Set
        End Property

        <Category("Appearance"), DefaultValue(""), Bindable(True), _
        DesignerSerializationVisibility(DesignerSerializationVisibility.Visible), _
        Browsable(True), EditorBrowsable(EditorBrowsableState.Always)> _
        Public Overrides Property Text() As String
            Get
                Return MyBase.Text
            End Get
            Set(ByVal value As String)
                MyBase.Text = value
            End Set
        End Property

        Protected Overrides Sub OnTextChanged(ByVal e As EventArgs)
            MyBase.OnTextChanged(e)

            If Text <> "" AndAlso Height < 16 Then
                Height = 16
            End If

            Invalidate()
        End Sub

        Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
            MyBase.OnPaint(e)

            If Width > Height Then
                Dim textSize As SizeF
                Dim g As Graphics = e.Graphics

                If Text <> "" Then
                    textSize = e.Graphics.MeasureString(Text, Font)

                    If Enabled Then
                        Using b As Brush = New SolidBrush(ForeColor)
                            g.DrawString(Text, Font, b, 0, (Height - Font.Height) \ 2)
                        End Using
                    Else
                        Using b As Brush = New SolidBrush(SystemColors.GrayText)
                            g.DrawString(Text, Font, b, 0, (Height - Font.Height) \ 2)
                        End Using
                    End If
                End If

                If Enabled Then
                    If IsGradient Then
                        Dim rect As New Rectangle(CInt(textSize.Width), Height \ 2, Width - CInt(textSize.Width), 1)
                        Dim p As New LinearGradientBrush(rect, SystemColors.ControlDarkDark, SystemColors.ControlLight, LinearGradientMode.Horizontal)
                        e.Graphics.FillRectangle(p, rect)
                        p.Dispose()
                    Else
                        Dim p As New Pen(SystemColors.ControlDark)
                        g.DrawLine(p, textSize.Width, Height \ 2, Width, Height \ 2)
                        p.Dispose()

                        g.DrawLine(Pens.White, textSize.Width, _
                             Height \ 2 + 1, Width, Height \ 2 + 1)
                    End If
                Else
                    Dim p As New Pen(SystemColors.InactiveBorder)
                    g.DrawLine(p, textSize.Width, Height \ 2, Width, Height \ 2)
                    p.Dispose()
                End If
            Else
                e.Graphics.DrawLine(New Pen(SystemColors.ControlDark), _
                      Width \ 2 - 1, 0, Width \ 2 - 1, Height)

                e.Graphics.DrawLine(Pens.White, _
                    Width \ 2, 0, Width \ 2, Height)
            End If
        End Sub
    End Class

    <ToolboxItem(False)> _
    Public Class WizardPageBase
        Inherits UserControl

        Private TitleFont As New Font(Font.FontFamily, Font.Size + 2, FontStyle.Bold)

        Public Sub New()
            Padding = New Padding(20)
            Size = New Size(500, 200)
            SetStyle(ControlStyles.OptimizedDoubleBuffer Or ControlStyles.AllPaintingInWmPaint, True)
            Padding = New Padding(20, CInt(MainImage.Height + 8), 20, 20)
        End Sub

        Private TitleValue As String

        <Category("WizardPage")> _
        Public Property Title() As String
            Get
                Return TitleValue
            End Get
            Set(ByVal Value As String)
                TitleValue = Value

                Dim padTop = CInt(MainImage.Height + TitleFont.Height + 8)

                If Padding.Top <> padTop Then
                    Padding = New Padding(20, padTop, 20, 20)
                End If

                Invalidate()
            End Set
        End Property

        Private DescriptionValue As String

        <Category("WizardPage")> _
        Public Property Description() As String
            Get
                Return DescriptionValue
            End Get
            Set(ByVal Value As String)
                DescriptionValue = Value

                Using g As Graphics = CreateGraphics()
                    Dim descSize As SizeF = g.MeasureString(Description, Font, Width - Padding.Left * 2 - 4)
                    Dim padTop = CInt(MainImage.Height + TitleFont.Height + 8 + descSize.Height + 8)

                    If Padding.Top <> padTop Then
                        Padding = New Padding(20, padTop, 20, 20)
                    End If
                End Using

                Invalidate()
            End Set
        End Property

        Private WizardFormValue As WizardForm

        Public Property WizardForm() As WizardForm
            Get
                Return WizardFormValue
            End Get
            Set(ByVal Value As WizardForm)
                WizardFormValue = Value
            End Set
        End Property

        Private Shared Shadows BackGroundImage As Image

        Private Shared MainImageValue As Image

        Public Shared ReadOnly Property MainImage() As Image
            Get
                If MainImageValue Is Nothing Then
                    MainImageValue = Image.FromFile(StaxApplication.MainImagePath)
                End If

                Return MainImageValue
            End Get
        End Property

        Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
            MyBase.OnPaint(e)

            If BackGroundImage Is Nothing Then
                BackGroundImage = Image.FromFile(StaxApplication.BackgroundImagePath)
            End If

            e.Graphics.DrawImage(BackGroundImage, 0, 0, Width, BackGroundImage.Height)
            e.Graphics.DrawImage(MainImage, 15, 5, MainImage.Width, MainImage.Height)

            If OK(Title) Then
                Using b As New SolidBrush(SystemColors.WindowText)
                    e.Graphics.DrawString(Title, TitleFont, b, Padding.Left + 2, MainImage.Height)
                End Using
            End If

            If OK(Description) Then
                Dim descSize = e.Graphics.MeasureString(Description, Font, Width - Padding.Left * 2 - 4)

                Using b As New SolidBrush(SystemColors.WindowText)
                    e.Graphics.DrawString(Description, Font, b, _
                        New RectangleF(Padding.Left + 2, MainImage.Height + TitleFont.Height + _
                        8, Width - Padding.Left * 2 - 4, descSize.Height))
                End Using
            End If

            Dim r, cr As Rectangle
            cr = ClientRectangle
            r = New Rectangle(cr.X + 1, cr.Y + 1, cr.Width - 2, cr.Height - 2)

            Dim c = ControlPaint.Light(Color.LightGray)

            Using p As New Pen(c, 2)
                e.Graphics.DrawRectangle(p, r)
            End Using
        End Sub
    End Class

    Public Class WizardPage(Of T)
        Inherits WizardPageBase

        Private Value As T
        Public ButtonHeight As Integer = 64

        Public Sub AddCommandButton(ByVal title As String, ByVal value As T)
            AddCommandButton(title, Nothing, value)
        End Sub

        Public Sub AddCommandButton( _
            ByVal title As String, _
            ByVal description As String, _
            ByVal value As T)

            AddCommandButton(title, Description, value, AddressOf CommandButtonClick)
        End Sub

        Private Sub AddCommandButton( _
            ByVal title As String, _
            ByVal description As String, _
            ByVal value As T, _
            ByVal handler As EventHandler)

            Dim cb As New CommandButton
            cb.Title = title
            cb.Description = description
            cb.Tag = value
            cb.Height = ButtonHeight
            AddHandler cb.Click, handler
            AddControl(cb)
        End Sub

        Public Sub AddDirectoryBrowserCommandButton( _
            ByVal title As String, _
            ByVal description As String)

            AddCommandButton(title, description, Nothing, _
                AddressOf DirectoryBrowserCommandButtonClick)
        End Sub

        Public Sub AddControl(ByVal c As Control)
            SuspendLayout()
            c.Top = Padding.Top + c.Margin.Top

            If Controls.Count > 0 Then
                Dim lc = Controls(Controls.Count - 1)
                c.Top = lc.Top + lc.Height + lc.Margin.Bottom + c.Margin.Top
            End If

            c.Anchor = AnchorStyles.Left Or AnchorStyles.Top Or AnchorStyles.Right
            c.Location = New Point(Padding.Left + c.Margin.Left, c.Top)
            c.Width = Width - Padding.Left - c.Margin.Left - Padding.Right - c.Margin.Right

            Dim clientHeight = c.Top + c.Height + c.Margin.Bottom + Padding.Bottom

            If ClientSize.Height < clientHeight Then
                ClientSize = New Size(ClientSize.Width, clientHeight)
            End If

            Controls.Add(c)
            ResumeLayout()
        End Sub

        Private Sub CommandButtonClick(ByVal o As Object, ByVal e As EventArgs)
            Value = DirectCast(DirectCast(o, CommandButton).Tag, T)
            WizardForm.NextPage()
        End Sub

        Private Sub DirectoryBrowserCommandButtonClick(ByVal o As Object, ByVal e As EventArgs)
            Using d As New FolderBrowserDialog
                d.Description = "Please select a directory."

                If d.ShowDialog = DialogResult.OK Then
                    Dim path As Object = d.SelectedPath
                    Value = DirectCast(path, T)
                    WizardForm.NextPage()
                End If
            End Using
        End Sub

        Public Function ShowPage() As T
            Using f As New WizardForm
                f.Text = Application.ProductName
                f.ClientSize = ClientSize
                f.AddPage(Me)
                f.ShowDialog()
                Return Value
            End Using
        End Function
    End Class

    <DesignerAttribute(GetType(CommandButtonDesigner))> _
    Public Class CommandButton
        Inherits Button

        Public Sub New()
            Anchor = AnchorStyles.Left Or AnchorStyles.Top Or AnchorStyles.Right
            Padding = New Padding(3)
        End Sub

        Private Function ShouldSerializeTitle() As Boolean
            Return OK(DescriptionValue)
        End Function

        Private TitleValue As String

        Public Property Title() As String
            Get
                Return TitleValue
            End Get
            Set(ByVal Value As String)
                TitleValue = Value
                Invalidate()
            End Set
        End Property

        Private Function ShouldSerializeDescription() As Boolean
            Return OK(DescriptionValue)
        End Function

        Private DescriptionValue As String

        Public Property Description() As String
            Get
                Return DescriptionValue
            End Get
            Set(ByVal Value As String)
                DescriptionValue = Value
                Invalidate()
            End Set
        End Property

        Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
            MyBase.OnPaint(e)

            Dim titleFont As New Font(Font.FontFamily, Font.Size + 2, FontStyle.Bold)
            Dim descriptionFont As New Font(Font.FontFamily, Font.Size, FontStyle.Regular)
            Dim pad = Padding.Left + 2
            Dim heightBoth As Integer
            Dim description = Me.Description

            If description Is Nothing Then
                description = ""
            End If

            Dim descriptionSize = e.Graphics.MeasureString(description, descriptionFont, ClientSize.Width - (pad * 2))

            Using b As New SolidBrush(SystemColors.WindowText)
                If description = "" Then
                    e.Graphics.DrawString(Title, titleFont, b, pad, (ClientSize.Height - titleFont.Height) \ 2)
                Else
                    heightBoth = CInt(titleFont.Height + descriptionSize.Height)
                    e.Graphics.DrawString(Title, titleFont, b, pad, (ClientSize.Height - heightBoth) \ 2)
                End If
            End Using

            Using b As New SolidBrush(SystemColors.WindowText)
                e.Graphics.DrawString(description, descriptionFont, b, New RectangleF(pad, ((ClientSize.Height - heightBoth) \ 2) + titleFont.Height, ClientSize.Width - (pad * 2), descriptionSize.Height))
            End Using
        End Sub
    End Class

    Public Enum TextBoxHintMode
        None
        Normal
        Hover
    End Enum

    Public Class ExtendedTextbox
        Inherits TextBox

        Private IsMouseEntered As Boolean

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property Name() As String
            Get
                Return MyBase.Name
            End Get
            Set(ByVal value As String)
                MyBase.Name = value
            End Set
        End Property

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property TabIndex() As Integer
            Get
                Return MyBase.TabIndex
            End Get
            Set(ByVal value As Integer)
                MyBase.TabIndex = value
            End Set
        End Property

        Private HintTextValue As String = ""

        <Category("ExtendedTextbox"), DefaultValue("")> _
        Public Property HintText() As String
            Get
                Return HintTextValue
            End Get
            Set(ByVal Value As String)
                HintTextValue = Value
                Invalidate()
            End Set
        End Property

        Private HintModeValue As TextBoxHintMode = TextBoxHintMode.None

        <Category("ExtendedTextbox"), DefaultValue(GetType(TextBoxHintMode), "None")> _
        Public Property HintMode() As TextBoxHintMode
            Get
                Return HintModeValue
            End Get
            Set(ByVal Value As TextBoxHintMode)
                HintModeValue = Value
                Invalidate()
            End Set
        End Property

        Protected Overrides Sub WndProc(ByRef msg As Message)
            MyBase.WndProc(msg)

            If msg.Msg = Win32.WM.PAINT Then
                If HintText <> "" AndAlso Text = "" AndAlso _
                    (HintMode = TextBoxHintMode.Normal OrElse _
                    IsMouseEntered) AndAlso Not Focused Then

                    Using g = CreateGraphics()
                        g.DrawString(HintText, Font, Brushes.Gray, ClientRectangle)
                    End Using
                End If
            End If
        End Sub

        Protected Overrides Sub OnMouseEnter(ByVal e As EventArgs)
            MyBase.OnMouseEnter(e)
            IsMouseEntered = True
            Refresh()
        End Sub

        Protected Overrides Sub OnMouseLeave(ByVal e As EventArgs)
            MyBase.OnMouseLeave(e)
            IsMouseEntered = False
            Refresh()
        End Sub

        Protected Overrides Sub OnGotFocus(ByVal e As EventArgs)
            MyBase.OnGotFocus(e)
            Refresh()
        End Sub

        Protected Overrides Sub OnLostFocus(ByVal e As EventArgs)
            MyBase.OnLostFocus(e)
            IsMouseEntered = False
            Refresh()
        End Sub
    End Class

    Public Class ExtendedPanel
        Inherits Panel

        Private IsAppThemed As Boolean = VisualStyleInformation.IsEnabledByUser AndAlso Application.RenderWithVisualStyles
        Private ShowNiceBorderValue As Boolean

        Public Sub New()
            SetStyle(ControlStyles.ResizeRedraw, True)
        End Sub

        <DefaultValue(False), Description("Nicer border if themes are enabled.")> _
        Public Property ShowNiceBorder() As Boolean
            Get
                Return ShowNiceBorderValue
            End Get
            Set(ByVal Value As Boolean)
                ShowNiceBorderValue = Value
                Invalidate()
            End Set
        End Property

        Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
            MyBase.OnPaint(e)

            If ShowNiceBorder AndAlso IsAppThemed Then
                ControlPaint.DrawBorder(e.Graphics, ClientRectangle, _
                    VisualStyleInformation.TextControlBorder, ButtonBorderStyle.Solid)
            End If
        End Sub
    End Class

    Public Class ExtendedCheckBox
        Inherits CheckBox

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property Name() As String
            Get
                Return MyBase.Name
            End Get
            Set(ByVal value As String)
                MyBase.Name = value
            End Set
        End Property

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property TabIndex() As Integer
            Get
                Return MyBase.TabIndex
            End Get
            Set(ByVal value As Integer)
                MyBase.TabIndex = value
            End Set
        End Property
    End Class

    Public Class ExtendedRichTextBox
        Inherits RichTextBox

        Private IsAppThemed As Boolean = VisualStyleInformation.IsEnabledByUser AndAlso Application.RenderWithVisualStyles

        Public BlockPaint As Boolean

        Protected Overrides Sub WndProc(ByRef m As Message)
            Select Case m.Msg
                Case WM.PAINT
                    If BlockPaint Then
                        Exit Sub
                    End If
            End Select

            MyBase.WndProc(m)

            Select Case m.Msg
                Case WM.PAINT
                    If IsAppThemed Then
                        Using g = CreateGraphics()
                            ControlPaint.DrawBorder(g, ClientRectangle, VisualStyleInformation.TextControlBorder, ButtonBorderStyle.Solid)
                        End Using
                    End If
                Case WM.VSCROLL
                    Refresh()
            End Select
        End Sub

        Protected Overrides ReadOnly Property CreateParams() As CreateParams
            Get
                Dim p = MyBase.CreateParams

                If IsAppThemed Then
                    p.ExStyle = p.ExStyle Xor CInt(WS.EX_CLIENTEDGE)
                End If

                Return p
            End Get
        End Property
    End Class

    Public Class ExtendedLabel
        Inherits Label

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property Name() As String
            Get
                Return MyBase.Name
            End Get
            Set(ByVal value As String)
                MyBase.Name = value
            End Set
        End Property

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property TabIndex() As Integer
            Get
                Return MyBase.TabIndex
            End Get
            Set(ByVal value As Integer)
                MyBase.TabIndex = value
            End Set
        End Property
    End Class

    Public Class ExtendedButton
        Inherits Button

        Shadows Event Click()

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property Name() As String
            Get
                Return MyBase.Name
            End Get
            Set(ByVal value As String)
                MyBase.Name = value
            End Set
        End Property

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property TabIndex() As Integer
            Get
                Return MyBase.TabIndex
            End Get
            Set(ByVal value As Integer)
                MyBase.TabIndex = value
            End Set
        End Property

        <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
        Shadows Property UseVisualStyleBackColor() As Boolean
            Get
                Return True
            End Get
            Set(ByVal value As Boolean)
            End Set
        End Property

        Private ShowMenuOnClickValue As Boolean

        <DefaultValue(False)> _
        Public Property ShowMenuOnClick() As Boolean
            Get
                Return ShowMenuOnClickValue
            End Get
            Set(ByVal Value As Boolean)
                ShowMenuOnClickValue = Value
            End Set
        End Property

        Private DropDownGlyphModeValue As DropDownGlyphMode

        <DefaultValue(GetType(DropDownGlyphMode), "None")> _
        Public Property DropDownGlyphMode() As DropDownGlyphMode
            Get
                Return DropDownGlyphModeValue
            End Get
            Set(ByVal Value As DropDownGlyphMode)
                DropDownGlyphModeValue = Value
                Invalidate()
            End Set
        End Property

        Protected Overrides Sub OnClick(ByVal e As EventArgs)
            MyBase.OnClick(e)
            RaiseEvent Click()

            If ShowMenuOnClick Then
                ContextMenuStrip.Show(Me, 0, Height)
            End If
        End Sub

        Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
            MyBase.OnPaint(e)

            Select Case DropDownGlyphMode
                Case UI.DropDownGlyphMode.Left
                    ButtonHelp.DrawDropDownGlyph(Me, e.Graphics, HorizontalAlignment.Left)
                Case UI.DropDownGlyphMode.Center
                    ButtonHelp.DrawDropDownGlyph(Me, e.Graphics, HorizontalAlignment.Center)
                Case UI.DropDownGlyphMode.Right
                    ButtonHelp.DrawDropDownGlyph(Me, e.Graphics, HorizontalAlignment.Right)
            End Select
        End Sub
    End Class

    Public Enum DropDownGlyphMode
        None
        Center
        Left
        Right
    End Enum

    Public Class MultiFolderTree
        Inherits ExtendedTreeView

        Public Sub New()
            AutoCollaps = True
            ExpandMode = TreeNodeExpandMode.Normal
        End Sub

        Public Sub Init()
            CheckBoxes = True
            Dim drives As New List(Of String)
            Dim volumes As New List(Of String)

            If Not DesignMode Then
                For Each i As DriveInfo In DriveInfo.GetDrives
                    If i.DriveType = DriveType.Fixed Then
                        drives.Add(i.RootDirectory.Name)

                        Try 'user reported crash
                            volumes.Add(i.VolumeLabel)
                        Catch ex As Exception
                            volumes.Add(i.Name)
                        End Try
                    End If
                Next

                Init(Nodes, drives.ToArray, volumes)

                For Each i As FolderTreeNode In Nodes
                    Init(i.Nodes, GetDirs(i.Path))
                Next
            End If
        End Sub

        Public Sub Init(ByVal nodes As TreeNodeCollection, ByVal directories As String(), Optional ByVal volumes As List(Of String) = Nothing)
            For i As Integer = 0 To directories.Length - 1
                Dim directory As String = directories(i)
                Dim di As New DirectoryInfo(directory)

                If directory.Length = 3 OrElse (di.Attributes And FileAttributes.Hidden) = 0 Then
                    Dim n As New FolderTreeNode
                    n.Path = directory

                    If volumes Is Nothing Then
                        n.Text = DirPath.GetName(directory)
                    Else
                        n.Text = volumes(i) + " (" + directory.Substring(0, 2) + ")"
                    End If

                    If n.Text = "" Then
                        n.Text = directory
                    End If

                    nodes.Add(n)
                End If
            Next
        End Sub

        <DebuggerNonUserCode()> _
        Private Function GetDirs(ByVal path As String) As String()
            Try
                Return Directory.GetDirectories(path)
            Catch ex As Exception
                Return New String() {}
            End Try
        End Function

        Protected Overrides Sub OnBeforeExpand(ByVal e As TreeViewCancelEventArgs)
            MyBase.OnBeforeExpand(e)

            Dim n As FolderTreeNode = DirectCast(e.Node, FolderTreeNode)

            For Each i As FolderTreeNode In n.Nodes
                If Not i.WasInitialized Then
                    Init(i.Nodes, GetDirs(i.Path))
                    i.WasInitialized = True
                End If
            Next
        End Sub

        Protected Overrides Sub OnAfterCheck(ByVal e As TreeViewEventArgs)
            Dim n As FolderTreeNode = DirectCast(e.Node, FolderTreeNode)

            If e.Node.Checked Then
                PathsValue.Add(n.Path)
            Else
                PathsValue.Remove(n.Path)
            End If

            MyBase.OnAfterCheck(e)
        End Sub

        Private PathsValue As New List(Of String)

        Public ReadOnly Property Paths() As List(Of String)
            Get
                Return PathsValue
            End Get
        End Property

        Protected Overrides Sub OnClick(ByVal e As EventArgs)
            MyBase.OnClick(e)

            Dim n = GetNodeAt(ClientMousePos)

            If Not n Is Nothing AndAlso (n.Nodes.Count = 0 OrElse n.IsExpanded) _
                AndAlso n.Bounds.Contains(ClientMousePos) Then

                n.Checked = Not n.Checked
            End If
        End Sub
    End Class

    Public Class FolderTreeNode
        Inherits TreeNode

        Public Path As String
        Public WasInitialized As Boolean
    End Class

    Public Class ExtendedListBox
        Inherits ListBox

        Private LastTick As Long
        Private KeyText As String = ""
        Private BlockOnSelectedIndexChanged As Boolean

        Public Sub UpdateSelection()
            If SelectedIndex > -1 Then
                BlockOnSelectedIndexChanged = True
                Items(SelectedIndex) = SelectedItem
                BlockOnSelectedIndexChanged = False

                If Sorted Then
                    Sorted = False
                    Sorted = True
                End If
            End If
        End Sub

        Protected Overrides Sub OnSelectedIndexChanged(ByVal e As EventArgs)
            If Not BlockOnSelectedIndexChanged Then
                MyBase.OnSelectedIndexChanged(e)
            End If
        End Sub

        Public Sub DeleteItem(ByVal text As String)
            If FindStringExact(text) > -1 Then
                Items.RemoveAt(FindStringExact(text))
            End If
        End Sub

        Public Sub RemoveSelection()
            If SelectedIndex > -1 Then
                If SelectionMode = Windows.Forms.SelectionMode.One Then
                    Dim index As Integer = SelectedIndex

                    If Items.Count - 1 > SelectedIndex Then
                        SelectedIndex += 1
                    Else
                        SelectedIndex -= 1
                    End If

                    Items.RemoveAt(index)
                Else
                    Dim iFirst As Integer = SelectedIndex

                    Dim indices(SelectedIndices.Count - 1) As Integer
                    SelectedIndices.CopyTo(indices, 0)

                    SelectedIndex = -1

                    For i As Integer = indices.Length - 1 To 0 Step -1
                        Items.RemoveAt(indices(i))
                    Next

                    If iFirst > Items.Count - 1 Then
                        SelectedIndex = Items.Count - 1
                    Else
                        SelectedIndex = iFirst
                    End If
                End If
            End If
        End Sub

        Public Function CanMoveUp() As Boolean
            Return SelectedIndices.Count > 0 AndAlso SelectedIndices(0) > 0
        End Function

        Public Function CanMoveDown() As Boolean
            Return SelectedIndices.Count > 0 AndAlso SelectedIndices(SelectedIndices.Count - 1) < Items.Count - 1
        End Function

        Public Sub MoveSelectionUp()
            If CanMoveUp() Then
                Dim iAbove As Integer = SelectedIndices(0) - 1

                If iAbove = -1 Then
                    Exit Sub
                End If

                Dim itemAbove As Object = Items(iAbove)
                Items.RemoveAt(iAbove)
                Dim iLastItem As Integer = SelectedIndices(SelectedIndices.Count - 1)
                Items.Insert(iLastItem + 1, itemAbove)
            End If
        End Sub

        Public Sub MoveSelectionDown()
            If CanMoveDown() Then
                Dim iBelow As Integer = SelectedIndices(SelectedIndices.Count - 1) + 1

                If iBelow >= Items.Count Then
                    Exit Sub
                End If

                Dim itemBelow As Object = Items(iBelow)
                Items.RemoveAt(iBelow)
                Dim iAbove As Integer = SelectedIndices(0) - 1
                Items.Insert(iAbove + 1, itemBelow)
            End If
        End Sub

        Public Sub KeyDownSequence(ByVal e As KeyEventArgs)
            If Environment.TickCount - LastTick > 1000 Then
                KeyText = Convert.ToChar(e.KeyValue).ToString()
            Else
                KeyText += Convert.ToChar(e.KeyValue).ToString()
            End If

            For i As Integer = 0 To Items.Count - 1
                If Items(i).ToString().ToLower().StartsWith(KeyText.ToLower) Then
                    Application.DoEvents()

                    SelectedIndex = -1
                    SelectedIndex = i

                    Exit For
                End If
            Next

            LastTick = Environment.TickCount
        End Sub

        Public Sub DrawNumbered(ByVal e As DrawItemEventArgs)
            If e.Index > -1 Then
                e.DrawBackground()
                e.DrawFocusRectangle()

                Dim sb As SolidBrush = New SolidBrush(ForeColor)
                Dim pf As PointF

                If e.Index > 998 Then
                    pf = New PointF(e.Bounds.Left + 40, e.Bounds.Top)
                Else
                    If e.Index > 98 Then
                        pf = New PointF(e.Bounds.Left + 30, e.Bounds.Top)
                    Else
                        pf = New PointF(e.Bounds.Left + 20, e.Bounds.Top)
                    End If
                End If

                e.Graphics.DrawString(Items(e.Index).ToString, Font, sb, pf)
                e.Graphics.DrawString((e.Index + 1).ToString, Font, sb, e.Bounds.Left, e.Bounds.Top)
            End If
        End Sub
    End Class

    Public Class BeforeShowControlEventArgs
        Inherits EventArgs

        Public Control As Control
        Public Position As Point
        Public Cancel As Boolean
    End Class

    Public Class ExtendedListView
        Inherits ListView

        Public Event Edited(ByVal value As Object, ByVal pos As Point)
        Public Event BeforeShowControl(ByVal e As BeforeShowControlEventArgs)

        Public Shadows Event DoubleClick()

        Private ColumnsDic As New Dictionary(Of Control, List(Of Integer))
        Private ControlsDic As New Dictionary(Of Integer, Control)
        Private CurrentPos As Point
        Private IsInit As Boolean
        Private MouseDownIndex As Integer
        Private InDoubleClickCheckHack As Boolean
        Private Const LVM_HITTEST As Integer = &H1000 + 18
        Private Const NM_DBLCLK As Integer = -3

        Public Sub New()
            SetStyle(ControlStyles.OptimizedDoubleBuffer, True)
        End Sub

        Private EditableValue As Boolean

        <DefaultValue(False)> _
        Public Property Editable() As Boolean
            Get
                Return EditableValue
            End Get
            Set(ByVal Value As Boolean)
                EditableValue = Value
            End Set
        End Property

        Private DoubleClickDoesCheckValue As Boolean = True

        <Category("Behavior"), DefaultValue(True)> _
        Public Property DoubleClickDoesCheck() As Boolean
            Get
                Return DoubleClickDoesCheckValue
            End Get
            Set(ByVal value As Boolean)
                DoubleClickDoesCheckValue = value
            End Set
        End Property

        Public Function CanMoveUp() As Boolean
            Return SelectedIndices.Count > 0 AndAlso SelectedIndices(0) > 0
        End Function

        Public Function CanMoveDown() As Boolean
            Return SelectedIndices.Count > 0 AndAlso SelectedIndices(SelectedIndices.Count - 1) < Items.Count - 1
        End Function

        Public Sub MoveSelectionUp()
            If CanMoveUp() Then
                Dim iAbove As Integer = SelectedIndices(0) - 1

                If iAbove = -1 Then
                    Exit Sub
                End If

                Dim itemAbove = Items(iAbove)
                Items.RemoveAt(iAbove)
                Dim iLastItem = SelectedIndices(SelectedIndices.Count - 1)
                Items.Insert(iLastItem + 1, itemAbove)
            End If
        End Sub

        Public Sub MoveSelectionDown()
            If CanMoveDown() Then
                Dim iBelow = SelectedIndices(SelectedIndices.Count - 1) + 1

                If iBelow >= Items.Count Then
                    Exit Sub
                End If

                Dim itemBelow = Items(iBelow)
                Items.RemoveAt(iBelow)
                Dim iAbove = SelectedIndices(0) - 1
                Items.Insert(iAbove + 1, itemBelow)
            End If
        End Sub

        Public Sub RemoveSelection()
            If SelectedItems.Count > 0 Then
                If Not MultiSelect Then
                    Dim index = SelectedIndices(0)

                    If Items.Count - 1 > index Then
                        Items(index + 1).Selected = True
                    Else
                        If index > 0 Then
                            Items(index - 1).Selected = True
                        End If
                    End If

                    Items.RemoveAt(index)
                Else
                    Dim iFirst = SelectedIndices(0)
                    Dim indices(SelectedIndices.Count - 1) As Integer
                    SelectedIndices.CopyTo(indices, 0)

                    For i = indices.Length - 1 To 0 Step -1
                        Items.RemoveAt(indices(i))
                    Next

                    If Items.Count > 0 Then
                        If iFirst > Items.Count - 1 Then
                            Items(Items.Count - 1).Selected = True
                        Else
                            Items(iFirst).Selected = True
                        End If
                    End If
                End If
            End If
        End Sub

        Public Sub AutoSizeColumns()
            For Each i As ColumnHeader In Columns
                Dim g = CreateGraphics()
                Dim s = g.MeasureString(i.Text, Font)
                i.AutoResize(ColumnHeaderAutoResizeStyle.ColumnContent)

                If i.Width < s.Width + 10 Then
                    i.Width = CInt(s.Width + 10)
                End If
            Next
        End Sub

        Private Sub HideControls()
            For Each i As Control In ControlsDic.Values
                i.Visible = False
                i.Enabled = False
            Next
        End Sub

        Protected Overrides Sub OnColumnReordered(ByVal e As ColumnReorderedEventArgs)
            If Editable Then
                Throw New NotImplementedException("Editing doesn't support collumn reorder.")
            End If

            MyBase.OnColumnReordered(e)
        End Sub

        Protected Overrides Sub OnColumnWidthChanging(ByVal e As ColumnWidthChangingEventArgs)
            HideControls()
            MyBase.OnColumnWidthChanging(e)
        End Sub

        'Unfortunately, the .NET ListView component automatically toggles the checked state of items
        'when you double click on them. I know that this is not the behavior of the underlying Win32 ListView control
        'so it has to be something in the WinForms code. At this point it's worth examing how this all works.
        'In a traditional C/C++ application, the ListView control sends WM_NOTIFY messages to the window that
        'is the parent of the ListView. This is typically a dialog box window. In WinForms, events are exposed directly
        'from the controls themselves.So internally WinForms will take the WM_NOTIFY message and reflect it back to the
        'child control and then the child control handles the message by firing events that you add your event handlers too.
        'This happens for other messages besides WM_NOTIFY - such as WM_COMMAND.

        'A few minutes with a program such as Spy++ will show you the message traffic.When you double click a ListView item
        'the underlying Win32 ListView sends a WM_NOTIFY message to the parent window (typically your Form). The WinForms
        'message handler for the parent window then reroutes the message back to the ListView by sending it a new message
        'WM_REFLECT + WM_NOTIFY. The WinForms ListView message handler then dispatches it. When the WinForms ListView sees
        'a NM_DBLCLK notification it then sends a message (LVM_HITTEST) to the Win32 ListView control asking where the click
        'occurred. If it was on an item, the WinForms ListView code will then toggle the checked state of the item.

        'Since none of this behavior is exposed via the properties of the ListView control we'll have to work around it using
        'less convenient means. The solution I came up with was to set a flag during the NM_DBLCLK notification that we're in
        'the midst of a double click notification and then we intercept the LVM_HITTEST call and return that no item was found.

        Protected Overrides Sub WndProc(ByRef m As Message)
            Select Case m.Msg
                Case WM.REFLECT + WM.NOTIFY
                    If Not DoubleClickDoesCheck AndAlso CheckBoxes Then
                        Dim s = DirectCast(Marshal.PtrToStructure(m.LParam, GetType(NMHDR)), NMHDR)

                        If s.code = NM_DBLCLK Then
                            InDoubleClickCheckHack = True
                        End If
                    End If
                Case LVM_HITTEST
                    If InDoubleClickCheckHack Then
                        InDoubleClickCheckHack = False
                        m.Result = New System.IntPtr(-1)
                        Return
                    End If
                Case WM.VSCROLL, WM.HSCROLL
                    HideControls()
                Case WM.LBUTTONDBLCLK
                    OnDoubleClick()
            End Select

            MyBase.WndProc(m)
        End Sub

        Protected Overrides Sub OnMouseDown(ByVal e As MouseEventArgs)
            MyBase.OnMouseDown(e)

            If SelectedIndices.Count > 0 Then
                MouseDownIndex = SelectedIndices(0)
            End If
        End Sub

        Protected Overrides Sub OnMouseUp(ByVal e As MouseEventArgs)
            If Not DragActive AndAlso Editable AndAlso _
                e.Button = Windows.Forms.MouseButtons.Left AndAlso _
                SelectedIndices.Count > 0 AndAlso _
                SelectedIndices(0) = MouseDownIndex AndAlso _
                Not Control.ModifierKeys = Keys.Control Then

                Dim b = GetBounds(e.Location)

                If b <> Rectangle.Empty Then
                    CurrentPos = GetPos(GetMousePos)

                    For Each i In ControlsDic.Keys
                        If i = CurrentPos.X Then
                            Dim pos = b.Location
                            Dim c = ControlsDic(i)
                            Dim args = New BeforeShowControlEventArgs()
                            args.Position = CurrentPos
                            args.Control = c
                            OnBeforeShowControl(args)

                            If Not args.Cancel Then
                                Dim offset = (c.Height - b.Height) \ 2
                                c.Location = New Point(pos.X, pos.Y - offset)
                                c.Size = b.Size
                                c.Enabled = True
                                IsInit = True
                                c.Text = GetText(CurrentPos)
                                IsInit = False
                                c.Visible = True
                                c.Focus()
                            End If

                            Exit For
                        End If
                    Next
                End If
            End If

            DragActive = False

            MyBase.OnMouseUp(e)
        End Sub

        Protected Overridable Sub OnEdited(ByVal value As Object, ByVal pos As Point)
            RaiseEvent Edited(value, pos)
        End Sub

        Protected Overridable Sub OnBeforeShowControl(ByVal e As BeforeShowControlEventArgs)
            RaiseEvent BeforeShowControl(e)
        End Sub

        Private Function GetBounds(ByVal mousePos As Point) As Rectangle
            Dim x, y, w, h, columnLeft, checkLength As Integer

            For Each i As ColumnHeader In Columns
                If i.Index = 0 AndAlso CheckBoxes Then
                    checkLength = 20
                Else
                    checkLength = 0
                End If

                If mousePos.X >= columnLeft + checkLength AndAlso _
                    mousePos.X <= columnLeft + i.Width Then

                    x = columnLeft + checkLength
                    w = i.Width - checkLength
                    Exit For
                End If

                columnLeft += i.Width
            Next

            For Each i As ListViewItem In Items
                Dim bounds As Rectangle = i.GetBounds(ItemBoundsPortion.Entire)
                If mousePos.Y >= bounds.Top AndAlso mousePos.Y <= bounds.Bottom Then
                    y = bounds.Top
                    h = bounds.Height
                End If
            Next

            If w = 0 OrElse h = 0 Then
                Return Rectangle.Empty
            Else
                Return New Rectangle(x, y, w, h)
            End If
        End Function

        Private Function GetPos(ByVal mousePos As Point) As Point
            Dim x, y, checkLength, columnLeft As Integer

            For Each i As ColumnHeader In Columns
                If i.Index = 0 AndAlso CheckBoxes Then
                    checkLength = 20
                Else
                    checkLength = 0
                End If

                If mousePos.X >= columnLeft + checkLength AndAlso _
                    mousePos.X <= columnLeft + i.Width Then

                    x = i.Index
                End If

                columnLeft += i.Width
            Next

            For Each i As ListViewItem In Items
                Dim bounds As Rectangle = i.GetBounds(ItemBoundsPortion.Entire)

                If mousePos.Y >= bounds.Top AndAlso mousePos.Y <= bounds.Bottom Then
                    y = i.Index
                End If
            Next

            Return New Point(x, y)
        End Function

        Private Function GetText(ByVal pos As Point) As String
            Dim item As ListViewItem = Items(pos.Y)
            Return item.SubItems(pos.X).Text
        End Function

        Protected Sub SetText(ByVal pos As Point, ByVal value As String)
            Dim item = Items(pos.Y)
            item.SubItems(pos.X).Text = value
        End Sub

        Private Sub ControlLostFocus(ByVal sender As Object, ByVal e As EventArgs)
            HideControls()
        End Sub

        Private Sub ControlValueChanged(ByVal sender As Object, ByVal e As EventArgs)
            If Not IsInit Then
                Dim c = DirectCast(sender, Control)
                SetText(CurrentPos, c.Text)

                Dim value As Object

                If TypeOf c Is ComboBox Then
                    Dim cb = DirectCast(c, ComboBox)

                    If cb.SelectedItem Is Nothing Then
                        value = cb.Text
                    Else
                        value = cb.SelectedItem
                    End If
                Else
                    value = c.Text
                End If

                OnEdited(value, CurrentPos)
            End If
        End Sub

        Public Sub AddControl(ByVal c As Control, ByVal columns As Integer())
            c.Visible = False
            c.Enabled = False

            AddHandler c.TextChanged, AddressOf ControlValueChanged
            AddHandler c.LostFocus, AddressOf ControlLostFocus

            Controls.Add(c)
            ColumnsDic(c) = New List(Of Integer)(columns)

            For Each i In columns
                ControlsDic(i) = c
            Next
        End Sub

        Public Sub AddTextBox(ByVal ParamArray columns As Integer())
            AddControl(New ListViewTextBox, columns)
        End Sub

        'enter won't fire when form has accept button set
        Private Class ListViewTextBox
            Inherits TextBox

            Protected Overrides Function ProcessCmdKey(ByRef msg As Message, ByVal keyData As Keys) As Boolean
                If keyData = Keys.Enter Then
                    Visible = False
                    Enabled = False
                    Return True
                End If

                Return MyBase.ProcessCmdKey(msg, keyData)
            End Function
        End Class

        'enter won't fire when form has accept button set
        Private Class ComboBoxEx
            Inherits ComboBox

            Protected Overrides Function ProcessCmdKey(ByRef msg As Message, ByVal keyData As Keys) As Boolean
                If keyData = Keys.Enter Then
                    Visible = False
                    Enabled = False
                    Return True
                End If

                Return MyBase.ProcessCmdKey(msg, keyData)
            End Function
        End Class

        Public Function AddComboBox(ByVal items As Object(), ByVal ParamArray columns As Integer()) As ComboBox
            Dim c As New ComboBoxEx
            c.MaxDropDownItems = 20
            c.Items.AddRange(items)
            AddHandler c.SelectedIndexChanged, AddressOf ControlValueChanged
            AddControl(c, columns)
            Return c
        End Function

        Protected Overridable Shadows Sub OnDoubleClick()
            RaiseEvent DoubleClick()
        End Sub

        Protected Overrides Sub OnDragEnter(ByVal e As DragEventArgs)
            e.Effect = DragDropEffects.Move
            MyBase.OnDragEnter(e)
        End Sub

        Protected Overrides Sub OnItemDrag(ByVal e As ItemDragEventArgs)
            If SelectedItems.Count > 1 Then
                Throw New NotImplementedException("multiselect drag is not implemented")
            End If

            DoDragDrop(e.Item, DragDropEffects.Move)
            DragActive = True
            MyBase.OnItemDrag(e)
        End Sub

        Private Function GetMousePos() As Point
            Return PointToClient(Control.MousePosition)
        End Function

        Private LastDragOverPos As Point
        Private LastDrawPos As Integer
        Private DragActive As Boolean

        'OnMouseMove doesn't work while dragging
        Protected Overrides Sub OnDragOver(ByVal e As DragEventArgs)
            If Control.MousePosition <> LastDragOverPos Then
                Dim mousePos = GetMousePos()
                Dim bounds = GetBounds(mousePos)

                If Not e.Data.GetDataPresent(GetType(ListViewItem)) Then
                    e.Effect = DragDropEffects.None
                ElseIf Not bounds = Rectangle.Empty Then
                    e.Effect = DragDropEffects.Move
                    Dim y As Integer

                    If Math.Abs(mousePos.Y - bounds.Top) < _
                        Math.Abs(mousePos.Y - bounds.Bottom) Then

                        y = bounds.Top
                    Else
                        y = bounds.Bottom
                    End If

                    If y <> LastDrawPos Then
                        LastDrawPos = y
                        Refresh()

                        Using g As Graphics = CreateGraphics()
                            g.DrawLine(Pens.Black, 0, y, Width, y)
                        End Using
                    End If
                Else
                    LastDrawPos = -1
                    e.Effect = DragDropEffects.None
                    Refresh()
                End If
            End If

            LastDragOverPos = Control.MousePosition
            MyBase.OnDragOver(e)
        End Sub

        Protected Overrides Sub OnDragDrop(ByVal e As DragEventArgs)
            Dim item = DirectCast(e.Data.GetData(GetType(ListViewItem)), ListViewItem)
            Dim mousePos As Point = GetMousePos()
            Dim p As Point = GetPos(mousePos)
            Dim r As Rectangle = GetBounds(mousePos)
            Dim index As Integer

            If Math.Abs(mousePos.Y - r.Top) < _
                Math.Abs(mousePos.Y - r.Bottom) Then

                index = p.Y
            Else
                index = p.Y + 1
            End If

            If index > item.Index Then
                index -= 1
            End If

            item.Remove()
            Items.Insert(index, item)
            MyBase.OnDragDrop(e)
        End Sub
    End Class

    Public Class ExtendedPropertyGrid
        Inherits PropertyGrid

        Private Grid As Reflector
        Private GridView As Reflector
        Private GridViewValue As Control
        Private CheckBoxes As New List(Of GridCheckBox)

        Public Sub New()
            ToolbarVisible = False 'GridView is not ready when PropertySortChanged happens so Init fails
        End Sub

        Private Class GridCheckBox
            Inherits CheckBox

            Public GridEntry As Reflector
            Private IsLoadComplete As Boolean

            Public Sub New(ByVal gridEntry As Reflector)
                Me.GridEntry = gridEntry

                If gridEntry.Invoke("GetPropertyTextValue").ToString = "" Then
                    CheckState = Windows.Forms.CheckState.Indeterminate
                Else
                    Checked = gridEntry.Invoke("Value").ToBool
                End If

                IsLoadComplete = True
            End Sub

            Protected Overrides Sub OnCheckedChanged(ByVal e As EventArgs)
                MyBase.OnCheckedChanged(e)

                If IsLoadComplete Then
                    GridEntry.Invoke("Select")
                    GridEntry.Invoke("GridEntryHost").Invoke("CommitValue", GridEntry.Value, Checked)
                End If
            End Sub
        End Class

        Public Sub Init()
            If Grid Is Nothing Then
                Grid = New Reflector(Me, GetType(PropertyGrid))
                GridView = Grid.Invoke("gridView")
                GridViewValue = GridView.Cast(Of Control)()
                AddHandler GridViewValue.Invalidated, AddressOf GridViewInvalidated
            End If

            Dim gridEntries = GridView.Invoke("allGridEntries")
            Dim gridEntriesValue As ICollection = gridEntries.Cast(Of ICollection)()

            For Each i In CheckBoxes
                GridViewValue.Controls.Remove(i)
                i.Dispose()
            Next

            If Not gridEntriesValue Is Nothing Then
                For Each i As Object In gridEntriesValue
                    Dim gridEntry As New Reflector(i)
                    Dim propertyTypeValue = gridEntry.Invoke("PropertyType").Cast(Of Type)()

                    If propertyTypeValue Is GetType(Boolean) Then
                        Dim c As New GridCheckBox(gridEntry)
                        CheckBoxes.Add(c)
                        Dim rect = GetRectangle(gridEntry)
                        c.Location = rect.Location
                        c.Size = rect.Size
                        GridViewValue.Controls.Add(c)
                    End If
                Next
            End If
        End Sub

        Private Sub GridViewInvalidated(ByVal sender As Object, ByVal e As InvalidateEventArgs)
            If Not GridView Is Nothing Then
                For Each i As GridCheckBox In CheckBoxes
                    i.Visible = i.GridEntry.Invoke("ParentGridEntry").Invoke("Expanded").ToBool
                    Dim rect = GetRectangle(i.GridEntry)

                    If rect.Left <> i.Left OrElse _
                        rect.Top <> i.Top Then

                        i.Width = rect.Width
                        i.Left = rect.Left
                        i.Top = rect.Top
                    End If

                    i.BringToFront()
                Next
            End If
        End Sub

        Public Shadows Property SelectedObject() As Object
            Get
                Return MyBase.SelectedObject
            End Get
            Set(ByVal Value As Object)
                MyBase.SelectedObject = Value
                Init()
            End Set
        End Property

        Public Shadows Property SelectedObjects() As Object()
            Get
                Return MyBase.SelectedObjects
            End Get
            Set(ByVal Value As Object())
                MyBase.SelectedObjects = Value
                Init()
            End Set
        End Property

        Private Function GetRectangle(ByVal gridEntry As Reflector) As Rectangle
            Dim row = GridView.Invoke("GetRowFromGridEntry", gridEntry.Value).ToInt
            Return GridView.Invoke("GetRectangle", row, 2).Cast(Of Rectangle)()
        End Function

        Protected Overrides Sub OnSelectedGridItemChanged(ByVal e As SelectedGridItemChangedEventArgs)
            MyBase.OnSelectedGridItemChanged(e)

            Dim help = e.NewSelection.PropertyDescriptor.Description

            If help <> "" Then
                HelpVisible = True

                Dim lines = CInt(Math.Ceiling(CreateGraphics.MeasureString( _
                    help, Font, Width).Height / Font.Height))

                Dim r As New Reflector(Me, GetType(PropertyGrid))
                Dim doc = r.Invoke("doccomment")
                doc.Invoke("Lines", lines + 1)
                doc.Invoke("userSized", True)
                r.Invoke("OnLayoutInternal", False)
            Else
                HelpVisible = False
            End If
        End Sub
    End Class

    <DefaultEvent("LinkClick")> _
    Public Class LinkGroupBox
        Inherits GroupBox

        WithEvents Label As New LinkLabel
        Event LinkClick()

        Public Sub New()
            Label.Left = 8
            Label.AutoSize = True
            Controls.Add(Label)
        End Sub

        Public Overrides Property Text() As String
            Get
                Return Label.Text
            End Get
            Set(ByVal value As String)
                Label.Text = value
            End Set
        End Property

        Private Sub Label_Click() Handles Label.Click
            RaiseEvent LinkClick()
        End Sub
    End Class

    Public Class ExtendedCheckedListBox
        Inherits CheckedListBox

        Private RemoveButtonValue As Button

        <DefaultValue(CStr(Nothing))> _
        Public Property RemoveButton() As Button
            Get
                Return RemoveButtonValue
            End Get
            Set(ByVal value As Button)
                RemoveButtonValue = value
                AddClickHandler(value)
            End Set
        End Property

        Private UpButtonValue As Button

        <DefaultValue(CStr(Nothing))> _
        Public Property UpButton() As Button
            Get
                Return UpButtonValue
            End Get
            Set(ByVal value As Button)
                UpButtonValue = value
                AddClickHandler(value)
            End Set
        End Property

        Private DownButtonValue As Button

        <DefaultValue(CStr(Nothing))> _
        Public Property DownButton() As Button
            Get
                Return DownButtonValue
            End Get
            Set(ByVal value As Button)
                DownButtonValue = value
                AddClickHandler(value)
            End Set
        End Property

        Private Sub AddClickHandler(ByVal b As Button)
            If Not DesignMode Then
                AddHandler b.Click, AddressOf ButtonClick
            End If
        End Sub

        Private Sub ButtonClick(ByVal s As Object, ByVal e As EventArgs)
            If s Is RemoveButton Then
                RemoveSelection()
            ElseIf s Is UpButton Then
                MoveSelectedItemUp()
            ElseIf s Is DownButton Then
                MoveSelectedItemDown()
            End If
        End Sub

        Public Sub MoveSelectedItemUp()
            Dim i = SelectedIndex

            If i > 0 Then
                Dim state = GetItemChecked(i)
                Items.Insert(i - 1, SelectedItem)
                Items.RemoveAt(i + 1)
                SelectedIndex = i - 1
                SetItemChecked(SelectedIndex, state)
            End If
        End Sub

        Public Sub MoveSelectedItemDown()
            Dim i = SelectedIndex

            If i < Items.Count - 1 Then
                Dim state = GetItemChecked(i)

                Items.Insert(i + 2, SelectedItem)
                Items.RemoveAt(i)
                SelectedIndex = i + 1
                SetItemChecked(SelectedIndex, state)
            End If
        End Sub

        Public Sub RemoveSelection()
            Dim i = SelectedIndex
            Items.Remove(SelectedItem)

            If Items.Count > i Then
                SelectedIndex = i
            Else
                If Items.Count > 0 Then
                    SelectedIndex = Items.Count - 1
                End If
            End If
        End Sub

        Protected Overrides Sub OnSelectedIndexChanged(ByVal e As EventArgs)
            MyBase.OnSelectedIndexChanged(e)
            RemoveButton.Enabled = Not SelectedItem Is Nothing
            UpButton.Enabled = SelectedIndex > 0
            DownButton.Enabled = SelectedIndex < Items.Count - 1
        End Sub
    End Class
End Namespace