Imports System.Reflection
Imports System.ComponentModel

Imports Stax
Imports Stax.UI
Imports Stax.ComponentModel

Namespace UI
    <Serializable()> _
    Public MustInherit Class Criteria
        Public MustOverride Function Eval() As Boolean
        Public MustOverride Property ValueString() As String
        Public MustOverride Property PropertyString() As String
        Public MustOverride ReadOnly Property ConditionNames() As String()
        Public MustOverride Property ConditionName() As String

        Private NameValue As String

        Public Property Name() As String
            Get
                Return NameValue
            End Get
            Set(ByVal Value As String)
                NameValue = Value
            End Set
        End Property

        Private MacroValue As String

        Public Property Macro() As String
            Get
                Return MacroValue
            End Get
            Set(ByVal Value As String)
                MacroValue = Value
            End Set
        End Property

        Public Shared Function Create(ByVal t As Type) As Criteria
            If t Is GetType(String) Then
                Return New StringCriteria
            ElseIf t Is GetType(Integer) Then
                Return New IntCriteria
            ElseIf t Is GetType(Boolean) Then
                Return New BoolCriteria
            End If

            Throw New Exception("failed to create criteria")
        End Function

        Public Overrides Function ToString() As String
            Return Name
        End Function
    End Class

    <Serializable()> _
    Public MustInherit Class GenericCriteria(Of TCondition, TType)
        Inherits Criteria

        Private PropertyValueValue As TType

        Public Property PropertyValue() As TType
            Get
                Return PropertyValueValue
            End Get
            Set(ByVal Value As TType)
                PropertyValueValue = Value
            End Set
        End Property

        Private ValueValue As TType

        Public Property Value() As TType
            Get
                Return ValueValue
            End Get
            Set(ByVal Value As TType)
                ValueValue = Value
            End Set
        End Property

        Private ConditionValue As TCondition

        Public Property Condition() As TCondition
            Get
                Return ConditionValue
            End Get
            Set(ByVal Value As TCondition)
                ConditionValue = Value
            End Set
        End Property

        Public Overrides ReadOnly Property ConditionNames() As String()
            Get
                Return DisplayInfoAttribute.GetNamesForEnum(Of TCondition)()
            End Get
        End Property

        Public Overrides Property ConditionName() As String
            Get
                Return DisplayInfoAttribute.GetValueForEnum(Of TCondition)(Condition)
            End Get
            Set(ByVal value As String)
                For Each i As TCondition In System.Enum.GetValues(GetType(TCondition))
                    If DisplayInfoAttribute.GetValueForEnum(Of TCondition)(i) = value Then
                        Condition = i
                    End If
                Next
            End Set
        End Property
    End Class

    <Serializable()> _
    Public Class IntCriteria
        Inherits GenericCriteria(Of IntCondition, Integer)

        Public Overrides Function Eval() As Boolean
            Select Case Condition
                Case IntCondition.Is
                    Return PropertyValue = Value
                Case IntCondition.IsNot
                    Return PropertyValue <> Value
                Case IntCondition.IsHigherThan
                    Return PropertyValue > Value
                Case IntCondition.IsLowerThan
                    Return PropertyValue < Value
            End Select
        End Function

        Public Overrides Property ValueString() As String
            Get
                Return Value.ToString
            End Get
            Set(ByVal value As String)
                Me.Value = value.ToInt(Me.Value)
            End Set
        End Property

        Public Overrides Property PropertyString() As String
            Get
                Return PropertyValue.ToString()
            End Get
            Set(ByVal value As String)
                PropertyValue = value.ToInt(PropertyValue)
            End Set
        End Property
    End Class

    <Serializable()> _
    Public Class StringCriteria
        Inherits GenericCriteria(Of StringCondition, String)

        Public Overrides Function Eval() As Boolean
            Select Case Condition
                Case StringCondition.Contains
                    Return PropertyValue.Contains(Value)
                Case StringCondition.DoesntContain
                    Return Not PropertyValue.Contains(Value)
                Case StringCondition.Is
                    Return PropertyValue = Value
                Case StringCondition.IsNot
                    Return PropertyValue <> Value
            End Select
        End Function

        Public Overrides Property ValueString() As String
            Get
                If Value Is Nothing Then
                    Value = ""
                End If

                Return Value
            End Get
            Set(ByVal value As String)
                Me.Value = value
            End Set
        End Property

        Public Overrides Property PropertyString() As String
            Get
                If PropertyValue Is Nothing Then
                    PropertyValue = ""
                End If

                Return Value
            End Get
            Set(ByVal value As String)
                PropertyValue = value
            End Set
        End Property
    End Class

    <Serializable()> _
    Public Class BoolCriteria
        Inherits GenericCriteria(Of BoolCondition, Boolean)

        Public Overrides Function Eval() As Boolean
            Select Case Condition
                Case BoolCondition.Is
                    Return PropertyValue = Value
                Case BoolCondition.IsNot
                    Return PropertyValue <> Value
                Case Else
                    Throw New NotImplementedException
            End Select
        End Function

        Public Overrides Property ValueString() As String
            Get
                Return Value.ToString
            End Get
            Set(ByVal value As String)
                If Not value Is Nothing AndAlso value.ToUpper = "TRUE" Then
                    Me.Value = True
                Else
                    Me.Value = False
                End If
            End Set
        End Property

        Public Overrides Property PropertyString() As String
            Get
                Return Value.ToString
            End Get
            Set(ByVal value As String)
                If Not value Is Nothing AndAlso value.ToUpper = "TRUE" Then
                    PropertyValue = True
                Else
                    PropertyValue = False
                End If
            End Set
        End Property
    End Class

    Public Enum IntCondition
        [Is]
        <DisplayInfo("Is Not")> [IsNot]
        <DisplayInfo("Is Higher Than")> IsHigherThan
        <DisplayInfo("Is Lower Than")> IsLowerThan
    End Enum

    Public Enum StringCondition
        [Is]
        <DisplayInfo("Is Not")> [IsNot]
        Contains
        <DisplayInfo("Contains Not")> DoesntContain
    End Enum

    Public Enum BoolCondition
        [Is]
        <DisplayInfo("Is Not")> [IsNot]
    End Enum
End Namespace