Imports System.Reflection
Imports System.ComponentModel
Imports System.Text
Imports System.Drawing.Design
Imports System.Runtime.InteropServices
Imports System.IO

Imports Stax.DynamicObject
Imports Stax.Win32
Imports Stax.ComponentModel
Imports Stax.AppFramework

Namespace UI
    Public Class ListBag(Of T)
        Public Sub New(ByVal text As String, ByVal value As T)
            Me.Text = text
            Me.Value = value
        End Sub

        Private TextValue As String

        Public Property Text() As String
            Get
                Return TextValue
            End Get
            Set(ByVal Value As String)
                TextValue = Value
            End Set
        End Property

        Private ValueValue As T

        Public Property Value() As T
            Get
                Return ValueValue
            End Get
            Set(ByVal Value As T)
                ValueValue = Value
            End Set
        End Property

        Public Shared Sub SelectItem(ByVal cb As ComboBox, ByVal value As T)
            Dim selectItem As Object = Nothing

            For Each i As ListBag(Of T) In cb.Items
                If i.Value.Equals(value) Then
                    selectItem = i
                End If
            Next

            If Not selectItem Is Nothing Then
                cb.SelectedItem = selectItem
            End If
        End Sub

        Public Shared Function GetValue(ByVal cb As ComboBox) As T
            Return DirectCast(DirectCast(cb.SelectedItem, ListBag(Of T)).Value, T)
        End Function

        Public Shared Function GetBagsForEnumType() As ListBag(Of T)()
            Dim l As New List(Of ListBag(Of T))

            For Each i As T In System.Enum.GetValues(GetType(T))
                l.Add(New ListBag(Of T)(DisplayInfoAttribute.GetValueForEnum(Of T)(i), i))
            Next

            Return l.ToArray
        End Function

        Public Overrides Function ToString() As String
            Return Text
        End Function
    End Class

    Public Delegate Sub MessageDispatchCallback(ByRef m As Message, ByRef cancel As Boolean)

    Public Class MessageFiler
        Implements IMessageFilter

        Private Cancel As Boolean
        Private Callback As MessageDispatchCallback

        Public Sub New(ByVal callback As MessageDispatchCallback)
            Me.Callback = callback
        End Sub

        Public Function PreFilterMessage(ByRef m As Message) As Boolean Implements IMessageFilter.PreFilterMessage
            Cancel = False
            Callback(m, Cancel)

            If Cancel Then
                Return True
            End If
        End Function
    End Class

    Public Class FileDialogFilter
        Public Shared Function GetAllFiles() As String
            Return "All Files|*.*"
        End Function

        Public Shared Function GetLine(ByVal text As String, ByVal ParamArray extensions As String()) As String
            Return text + " (" + String.Join(", ", extensions) + ")|" + "*." + _
                String.Join(";*.", extensions)
        End Function
    End Class

    <Serializable()> _
    Public Class WindowPositions
        Private Sizes As New Dictionary(Of String, Size)
        Private Positions As New Dictionary(Of String, Point)
        Private WindowStates As New Dictionary(Of String, FormWindowState)

        Public Enum Mode
            None = 0
            WindowState = 2
            Size = 4
        End Enum

        Public Sub Save(ByVal f As Form)
            SavePosition(f)
            SaveSize(f)
            SaveWindowState(f)
        End Sub

        Private Sub SavePosition(ByVal f As Form)
            If f.WindowState = FormWindowState.Normal Then
                Positions(GetKey(f)) = f.Location
            End If
        End Sub

        Private Sub SaveSize(ByVal f As Form)
            If f.WindowState = FormWindowState.Normal Then
                Sizes(GetKey(f)) = f.Size
            End If
        End Sub

        Private Sub SaveWindowState(ByVal f As Form)
            WindowStates(GetKey(f)) = f.WindowState
        End Sub

        Private Sub RestorePositionInternal(ByVal f As Form)
            f.StartPosition = FormStartPosition.Manual

            If Positions.ContainsKey(GetKey(f)) Then
                Dim p = Positions(GetKey(f))

                If SystemInformation.WorkingArea.Contains(p) AndAlso _
                    SystemInformation.WorkingArea.Contains(p.X + f.Width, p.Y + f.Height) Then

                    f.Location = p
                Else
                    f.StartPosition = FormStartPosition.CenterParent
                End If
            Else
                f.StartPosition = FormStartPosition.CenterParent
            End If
        End Sub

        Private Sub RestoreSize(ByVal f As Form)
            If Sizes.ContainsKey(GetKey(f)) Then
                f.Size = Sizes(GetKey(f))
            End If
        End Sub

        Private Sub RestoreWindowState(ByVal f As Form)
            If WindowStates.ContainsKey(GetKey(f)) Then
                f.WindowState = WindowStates(GetKey(f))
            End If
        End Sub

        Public Sub Restore(ByVal f As Form)
            Restore(f, WindowPositions.Mode.WindowState Or WindowPositions.Mode.Size)
        End Sub

        Public Sub RestorePosition(ByVal f As Form)
            Dim value = f.Text

            If TypeOf f Is HelpForm Then
                value = "Help"
            End If

            If OK(StaxApplication.WindowPositionsCenterScreen) Then
                For Each i In StaxApplication.WindowPositionsCenterScreen
                    If value.StartsWith(i) OrElse i = "all" Then
                        f.StartPosition = FormStartPosition.CenterScreen
                        Exit For
                    End If
                Next
            End If

            If OK(StaxApplication.WindowPositionsRemembered) Then
                For Each i In StaxApplication.WindowPositionsRemembered
                    If value.StartsWith(i) OrElse i = "all" Then
                        RestorePositionInternal(f)
                        Exit For
                    End If
                Next
            End If
        End Sub

        Public Sub Restore(ByVal f As Form, ByVal m As Mode)
            If m <> Mode.None Then
                If (m And Mode.Size) = Mode.Size Then
                    RestoreSize(f)
                End If

                If (m And Mode.WindowState) = Mode.WindowState Then
                    RestoreWindowState(f)
                End If
            End If
        End Sub

        Private Function GetKey(ByVal f As Form) As String
            Return f.Name + f.GetType().FullName + f.Text
        End Function
    End Class

    Public Class ExtendedHelpEventArgs
        Inherits HelpEventArgs

        Public Sub New(ByVal mousePos As Point)
            MyBase.New(mousePos)
        End Sub

        Private HelpTitleValue As String

        Public Property HelpTitle() As String
            Get
                Return HelpTitleValue
            End Get
            Set(ByVal Value As String)
                HelpTitleValue = Value

                If Not HelpTitleValue Is Nothing Then
                    If HelpTitleValue.StartsWith(" "c) Then
                        HelpTitleValue = HelpTitleValue.TrimStart(" "c)
                    End If

                    If HelpTitleValue.EndsWith("...") Then
                        HelpTitleValue = HelpTitleValue.TrimEnd("."c)
                    End If

                    If HelpTitleValue.EndsWith(":") Then
                        HelpTitleValue = HelpTitleValue.TrimEnd(":"c)
                    End If
                End If
            End Set
        End Property

        Private HelpTextValue As String

        Public Property HelpText() As String
            Get
                Return HelpTextValue
            End Get
            Set(ByVal Value As String)
                HelpTextValue = Value
            End Set
        End Property

        Private HelpTopicValue As String

        Public Property HelpTopic() As String
            Get
                Return HelpTopicValue
            End Get
            Set(ByVal Value As String)
                HelpTopicValue = Value
            End Set
        End Property
    End Class

    Public Interface IHostedControl
        Sub ShowHelp()
        Property HelpID() As String
    End Interface

    Public Class OpenFileDialogEditor
        Inherits UITypeEditor

        Public Overloads Overrides Function EditValue(ByVal context As ITypeDescriptorContext, ByVal provider As IServiceProvider, ByVal value As Object) As Object
            Using f As New OpenFileDialog
                If f.ShowDialog = DialogResult.OK Then
                    Return f.FileName
                Else
                    Return value
                End If
            End Using
        End Function

        Public Overloads Overrides Function GetEditStyle(ByVal context As ITypeDescriptorContext) As UITypeEditorEditStyle
            Return UITypeEditorEditStyle.Modal
        End Function
    End Class

    Public Class StringEditor
        Inherits UITypeEditor

        Public Sub New()
        End Sub

        Public Overloads Overrides Function EditValue(ByVal context As ITypeDescriptorContext, ByVal provider As IServiceProvider, ByVal value As Object) As Object
            Dim f As New StringEditorForm
            f.tb.Text = DirectCast(value, String)

            If f.ShowDialog() = DialogResult.OK Then
                Return f.tb.Text.FixBreak
            Else
                Return value
            End If
        End Function

        Public Overloads Overrides Function GetEditStyle(ByVal context As ITypeDescriptorContext) As UITypeEditorEditStyle
            Return UITypeEditorEditStyle.Modal
        End Function
    End Class
End Namespace