Imports System.ComponentModel.Design.Serialization
Imports System.ComponentModel
Imports System.Drawing.Design
Imports System.Reflection
Imports System.Resources
Imports System.IO
Imports System.Xml
Imports System.Text
Imports System.Text.RegularExpressions

Imports Stax.AppFramework

Namespace UI
    <ProvideProperty("TipText", GetType(Control)), _
    ProvideProperty("TipName", GetType(Control))> _
    Public Class TipProvider
        Inherits Component
        Implements IExtenderProvider, ISupportInitialize, IDisposable

        Private ToolTip As ToolTip

        Public Sub New(ByVal cont As IContainer)
            cont.Add(Me)

            If Not DesignMode AndAlso StaxApplication.ShowToolTips Then
                ToolTip = New ToolTip(cont)
                ToolTip.AutomaticDelay = 1000
                ToolTip.AutoPopDelay = 10000
                ToolTip.InitialDelay = 1000
                ToolTip.ReshowDelay = 1000
            End If
        End Sub

        Private TipNamesValue As New Dictionary(Of Control, String)

        Public ReadOnly Property TipNames() As Dictionary(Of Control, String)
            Get
                Return TipNamesValue
            End Get
        End Property

        Private TipTextsValue As New Dictionary(Of Control, String)

        Public ReadOnly Property TipTexts() As Dictionary(Of Control, String)
            Get
                Return TipTextsValue
            End Get
        End Property

        Public Function CanExtend(ByVal obj As Object) As Boolean Implements IExtenderProvider.CanExtend
            Return TypeOf obj Is Control
        End Function

        <Category("TipProvider"), DefaultValue("")> _
        Public Function GetTipName(ByVal c As Control) As String
            If TipNames.ContainsKey(c) Then
                Return TipNames(c)
            End If

            Return ""
        End Function

        Public Sub SetTipName(ByVal c As Control, ByVal value As String)
            TipNames(c) = value
        End Sub

        <Category("TipProvider"), DefaultValue(""), _
        Editor(GetType(StringEditor), GetType(UITypeEditor))> _
        Public Function GetTipText(ByVal c As Control) As String
            If TipTexts.ContainsKey(c) Then
                Return TipTexts(c)
            End If

            Return ""
        End Function

        Public Sub SetTip(ByVal c As Control, ByVal value As String)
            SetTip(c, c.Text, value)
        End Sub

        Public Sub SetTip(ByVal c As Control, ByVal name As String, ByVal value As String)
            SetTipName(c, name)
            SetTipText(c, value)
            AddHandlers(c)
        End Sub

        Public Sub SetTipText(ByVal c As Control, ByVal value As String)
            TipTexts(c) = value
        End Sub

        Public Sub BeginInit() Implements ISupportInitialize.BeginInit
        End Sub

        Public Sub EndInit() Implements ISupportInitialize.EndInit
            For Each i In TipTexts.Keys
                AddHandlers(i)
            Next
        End Sub

        Private Sub AddHandlers(ByVal c As Control)
            If Not DesignMode Then
                AddHandler c.MouseDown, AddressOf TipMouseDown

                If StaxApplication.ShowToolTips Then
                    Dim s = GetTipText(c)

                    If Not s Is Nothing Then
                        s = s.TrimEnd("."c)

                        If s.Length > 80 AndAlso Not HasContextMenu(c) Then
                            s = "Please right-click for help"
                        ElseIf HelpDocument.MustConvert(s) Then
                            s = HelpDocument.ConvertMarkup(s, True)
                        End If

                        ToolTip.SetToolTip(c, s)
                    End If
                End If
            End If
        End Sub

        Private Sub TipMouseEnter(ByVal sender As Object, ByVal e As EventArgs)
            ToolTip.ToolTipTitle = GetTipName(DirectCast(sender, Control))
        End Sub

        Private Sub TipMouseDown(ByVal sender As Object, ByVal e As MouseEventArgs)
            If e.Button = MouseButtons.Right AndAlso _
                Not HasContextMenu(DirectCast(sender, Control)) Then

                ShowHelp(DirectCast(sender, Control), e.Location)
            End If
        End Sub

        Private Sub ShowHelp(ByVal c As Control, ByVal p As Point)
            Dim t = GetTip(c)
            Dim e As New ExtendedHelpEventArgs(p)

            If HelpDocument.MustConvert(t.Value) Then
                e.HelpText = HelpDocument.ConvertMarkup(t.Value, True)
            Else
                e.HelpText = t.Value
            End If

            If t.Name = "" Then
                e.HelpTitle = GetControlText(c)
            Else
                e.HelpTitle = t.Name
            End If

            StaxApplication.RaiseHelpRequested(e)
        End Sub

        Private Function GetTip(ByVal c As Control) As StringPair
            Dim t As New StringPair

            If TipNames.ContainsKey(c) Then
                t.Name = TipNames(c)
            End If

            If TipTexts.ContainsKey(c) Then
                t.Value = TipTexts(c)
            End If

            Return t
        End Function

        Private Function HasContextMenu(ByVal c As Control) As Boolean
            If TypeOf c Is TextBox Then Return True
            If TypeOf c Is RichTextBox Then Return True
            If TypeOf c Is NumericUpDown Then Return True

            If TypeOf c Is ComboBox AndAlso CType(c, ComboBox).DropDownStyle <> ComboBoxStyle.DropDownList Then
                Return True
            End If

            Dim pi = c.GetType.GetProperty("ContextMenuStrip")

            If Not pi Is Nothing Then
                If Not pi.GetValue(c, Nothing) Is Nothing Then
                    Return True
                End If
            End If

            Return False
        End Function

        Private Function GetControlText(ByVal c As Control) As String
            Dim ret = c.Text

            If ret.Contains(" ") Then
                ret = ret.Trim
            End If

            If ret.Contains("&") AndAlso Not ret.Contains(" & ") Then
                ret = ret.Replace("&", "")
            End If

            Return ret
        End Function

        Public Function GetTips() As StringPairList
            Dim list As New StringPairList
            Dim names As New List(Of String) 'to ignore double tips

            For Each i As Control In TipTexts.Keys
                If Not i.IsDisposed Then 'controls in background tabs are not visible!
                    Dim t As New StringPair
                    t.Value = TipTexts(i)

                    If TipNames.ContainsKey(i) Then
                        t.Name = TipNames(i)
                    Else
                        t.Name = GetControlText(i)
                    End If

                    If Not names.Contains(t.Name) Then
                        names.Add(t.Name)

                        If t.Name.EndsWith("...") Then
                            t.Name = t.Name.TrimEnd("."c)
                        End If

                        If t.Name.EndsWith(":") Then
                            t.Name = t.Name.TrimEnd(":"c)
                        End If

                        list.Add(t)
                    End If
                End If
            Next

            list.Sort()

            Return list
        End Function
    End Class

    Public Class PopUp
        Private TitleValue As String

        Public Property Title() As String
            Get
                Return TitleValue
            End Get
            Set(ByVal value As String)
                TitleValue = value
            End Set
        End Property

        Private TextValue As String

        Public Property Text() As String
            Get
                Return TextValue
            End Get
            Set(ByVal value As String)
                TextValue = value
            End Set
        End Property

        Private FontValue As Font

        Public Property Font() As Font
            Get
                Return FontValue
            End Get
            Set(ByVal Value As Font)
                FontValue = Value
            End Set
        End Property

        Private ImageValue As Image

        Public Property Image() As Image
            Get
                Return ImageValue
            End Get
            Set(ByVal value As Image)
                ImageValue = value
            End Set
        End Property

        Private CenterOwnerValue As Control

        Public Property CenterOwner() As Control
            Get
                Return CenterOwnerValue
            End Get
            Set(ByVal value As Control)
                CenterOwnerValue = value
            End Set
        End Property

        Public WriteOnly Property Tip() As StringPair
            Set(ByVal Value As StringPair)
                Title = Value.Name
                Text = Value.Value
            End Set
        End Property

        Public Sub Show()
            Using f As New PopUpForm()
                f.Title = Title
                f.TextValue = Text
                f.Image = Image
                f.CenterOwner = CenterOwner

                If Not Font Is Nothing Then
                    f.Font = Font
                End If

                f.ShowDialog()
            End Using
        End Sub

        Public Shared Sub ShowOnce(ByVal centerOwner As Control, ByVal text As String)
            If Not StaxApplication.StringList.Contains(text.MD5Hash) Then
                StaxApplication.StringList.Add(text.MD5Hash)

                Dim pop As New PopUp
                pop.CenterOwner = centerOwner
                pop.Font = New Font(Control.DefaultFont.FontFamily, 11)
                pop.Text = text
                pop.Show()
            End If
        End Sub
    End Class
End Namespace