Imports System.ComponentModel
Imports System.Drawing.Design
Imports System.Text
Imports System.Runtime.Serialization
Imports System.Reflection

Imports Stax
Imports Stax.ComponentModel
Imports Stax.UI

<Serializable()> _
Public Class AviSynthDocument
    Inherits Profile

    <NonSerialized()> Private Framerate As Single
    <NonSerialized()> Private Frames As Integer
    <NonSerialized()> Private Size As Size
    <NonSerialized()> Private FourCC As String = ""
    <NonSerialized()> Private ErrorMessage As String
    <NonSerialized()> Private LastSync As String

    Public Sub New()
        Me.New(Nothing)
    End Sub

    Public Sub New(ByVal name As String)
        MyBase.New(name, "Script Profile")
    End Sub

    Private FiltersValue As New List(Of AviSynthFilter)

    Public ReadOnly Property Filters() As List(Of AviSynthFilter)
        Get
            Return FiltersValue
        End Get
    End Property

    Private PathValue As String = ""

    Public Overridable Property Path() As String
        Get
            Return PathValue
        End Get
        Set(ByVal Value As String)
            PathValue = Value
        End Set
    End Property

    Public Overridable Function GetScript() As String
        Return GetScript(Nothing)
    End Function

    Public Overridable Function GetScript(ByVal skipCategory As String) As String
        Dim sb As New StringBuilder()

        For Each i As AviSynthFilter In Filters
            If i.Active Then
                If skipCategory Is Nothing OrElse i.Category <> skipCategory Then
                    sb.Append(i.Script + CrLf)
                End If
            End If
        Next

        Return sb.ToString
    End Function

    Public Sub Remove(ByVal category As String)
        While Not GetFilter(category) Is Nothing
            Filters.Remove(GetFilter(category))
        End While
    End Sub

    Public Sub SetFilter(ByVal category As String, ByVal name As String, ByVal script As String)
        For Each i In Filters
            If i.Category = category Then
                i.Name = name
                i.Script = script
                i.Active = True
                Exit For
            End If
        Next
    End Sub

    Public Sub InsertAfter(ByVal category As String, ByVal af As AviSynthFilter)
        Dim f = GetFilter(category)
        Filters.Insert(Filters.IndexOf(f) + 1, af)
    End Sub

    Public Function IsFilterActive(ByVal category As String) As Boolean
        Dim f = GetFilter(category)
        Return Not f Is Nothing AndAlso f.Active
    End Function

    Public Function GetFilter(ByVal category As String) As AviSynthFilter
        For Each i In Filters
            If i.Category = category Then
                Return i
            End If
        Next

        Return Nothing
    End Function

    Public Sub Synchronize()
        If Path <> "" Then
            Dim script = Macro.Solve(GetScript())
            Dim current = Path + script

            If Frames = 240 OrElse current <> LastSync Then
                If Directory.Exists(Filepath.GetDir(Path)) Then
                    script.WriteFile(Path)
                    Dim a As New AVIFile
                    a.Open(Path)
                    Framerate = a.FrameRate
                    Frames = a.FrameCount
                    FourCC = a.FourCC
                    Size = a.FrameSize
                    ErrorMessage = a.ErrorMessage
                    a.Close()
                    LastSync = current
                End If
            End If
        End If
    End Sub

    Public Function ContainsIgnoreCase(ByVal value As String) As Boolean
        For Each i In Filters
            If i.Active AndAlso i.Script.ToUpper.Contains(value.ToUpper) Then
                Return True
            End If
        Next

        Return False
    End Function

    Public Function GetFramerate() As Double
        Synchronize()
        Return Framerate
    End Function

    Public Function GetFourCC() As String
        Synchronize()
        Return FourCC
    End Function

    Public Function GetErrorMessage() As String
        Synchronize()
        Return ErrorMessage
    End Function

    Public Function GetSeconds() As Integer
        Return CInt(GetFrames() / GetFramerate())
    End Function

    Public Function GetFrames() As Integer
        Synchronize()
        Return Frames
    End Function

    Public Function GetSize() As Size
        Synchronize()
        Return Size
    End Function

    Public Shared Function GetDefaults() As List(Of TargetAviSynthDocument)
        Dim ret As New List(Of TargetAviSynthDocument)

        Dim mpeg2 As New TargetAviSynthDocument("Default Filter Setup")
        mpeg2.Filters.Add(New AviSynthFilter("Source", "Automatic", "", True))
        mpeg2.Filters.Add(New AviSynthFilter("Field", "Yadif", "LoadCplugin(""%application:Yadif%"") Yadif()", False))
        mpeg2.Filters.Add(New AviSynthFilter("Crop", "Crop Borders", "Crop(%crop_left%,%crop_top%,-%crop_right%,-%crop_bottom%)", True))
        mpeg2.Filters.Add(New AviSynthFilter("Noise", "Medium", "FluxSmoothST(4)", False))
        mpeg2.Filters.Add(New AviSynthFilter("Resize", "Neutral", "BicubicResize(%target_width%,%target_height%,0,0.5)", False))
        ret.Add(mpeg2)

        Return ret
    End Function

    Public Overrides Function GetInfoValues() As System.Collections.Generic.Dictionary(Of String, String)
        Dim ret As New Dictionary(Of String, String)

        For Each i As AviSynthFilter In Filters
            ret(i.Category) = i.Script
        Next

        Return ret
    End Function

    Public Overrides Function Edit() As DialogResult
        AviSynthListView.EditClick(AddressOf GetDocument)
        Return DialogResult.OK
    End Function

    Private Function GetDocument() As AviSynthDocument
        Return Me
    End Function
End Class

<Serializable()> _
Public Class TargetAviSynthDocument
    Inherits AviSynthDocument

    Public Sub New(ByVal name As String)
        MyBase.New(name)
        CanEditValue = True
    End Sub

    Public Overrides Property Path() As String
        Get
            If p.SourceFile = "" OrElse p.Name = "" Then
                Return ""
            End If

            Return g.TempDir + p.Name + ".avs"
        End Get
        Set(ByVal value As String)
        End Set
    End Property
End Class

<Serializable()> _
Public Class SourceAviSynthDocument
    Inherits AviSynthDocument

    Public Sub New()
        Me.New(Nothing)
    End Sub

    Public Sub New(ByVal name As String)
        MyBase.New(name)
    End Sub

    Public Overrides Property Path() As String
        Get
            If p.SourceFile = "" Then
                Return ""
            End If

            Return g.TempDir + p.Name + "_Source.avs"
        End Get
        Set(ByVal value As String)
        End Set
    End Property

    Public Overrides Function GetScript() As String
        Dim l As Integer
        Dim found As Boolean = False

        For Each i As AviSynthFilter In p.TargetAVSDoc.Filters
            If i.Active AndAlso (i.Category = "Source" OrElse _
                i.Category = "Plugin") Then

                found = True
                l += i.Script.Length + CrLf.Length
            End If
        Next

        Dim sb As New StringBuilder(l)

        For Each i As AviSynthFilter In p.TargetAVSDoc.Filters
            If i.Active AndAlso (i.Category = "Source" OrElse _
                i.Category = "Plugin") Then

                sb.Append(i.Script + CrLf)
            End If
        Next

        Return sb.ToString
    End Function
End Class

<Serializable()> _
Public Class AviSynthFilter
    Public Sub New()
        Me.new("???", "???", "???", True)
    End Sub

    Public Sub New(ByVal category As String, ByVal name As String, ByVal script As String, ByVal active As Boolean)
        Me.Name = name
        Me.Script = script
        Me.Category = category
        Me.Active = active
    End Sub

    Private ActiveValue As Boolean

    Public Property Active() As Boolean
        Get
            Return ActiveValue
        End Get
        Set(ByVal Value As Boolean)
            ActiveValue = Value
        End Set
    End Property

    Private CategoryValue As String

    Public Property Category() As String
        Get
            Return CategoryValue
        End Get
        Set(ByVal Value As String)
            CategoryValue = Value
        End Set
    End Property

    Private NameValue As String

    Public Property Name() As String
        Get
            Return NameValue
        End Get
        Set(ByVal Value As String)
            NameValue = Value
        End Set
    End Property

    Private ScriptValue As String

    Public Property Script() As String
        Get
            Return ScriptValue
        End Get
        Set(ByVal Value As String)
            ScriptValue = Value.FixBreak
        End Set
    End Property

    Public Function GetCopy() As AviSynthFilter
        Return New AviSynthFilter(Category, Name, Script, Active)
    End Function

    Public Overrides Function ToString() As String
        Return Name
    End Function
End Class

<Serializable()> _
Public Class AviSynthCategory
    Public Sub New(ByVal name As String)
        Me.Name = name
    End Sub

    Private NameValue As String

    Public Property Name() As String
        Get
            Return NameValue
        End Get
        Set(ByVal Value As String)
            NameValue = Value
        End Set
    End Property

    Private FitersValue As New ArrayList
    Private FitersNewValue As New List(Of AviSynthFilter)

    Public ReadOnly Property Filters() As List(Of AviSynthFilter)
        Get
            If FitersNewValue Is Nothing Then
                FitersNewValue = New List(Of AviSynthFilter)

                For Each i As AviSynthFilter In FitersValue
                    FitersNewValue.Add(i)
                Next
            End If

            Return FitersNewValue
        End Get
    End Property

    Public Overrides Function ToString() As String
        Return Name
    End Function

    Public Shared Function GetDefaults() As List(Of AviSynthCategory)
        Dim l As New List(Of AviSynthCategory)

        Dim src As New AviSynthCategory("Source")
        src.Filters.Add(New AviSynthFilter(src.Name, "Automatic", "", True))
        src.Filters.Add(New AviSynthFilter(src.Name, "MPEG-2", "MPEG2Source(""%source_file%"")", True))
        src.Filters.Add(New AviSynthFilter(src.Name, "MPEG-2 Deblocked", "MPEG2Source(""%source_file%"", CPU=6)", True))
        src.Filters.Add(New AviSynthFilter(src.Name, "AVC", "AVCSource(""%source_file%"")", True))
        src.Filters.Add(New AviSynthFilter(src.Name, "AVI", "AVISource(%source_files_comma%,audio=false)", True))
        src.Filters.Add(New AviSynthFilter(src.Name, "DirectShow", "DirectShowSource(""%source_file%"", audio=false, convertfps=true)", True))
        l.Add(src)

        Dim crop As New AviSynthCategory("Crop")
        crop.Filters.Add(New AviSynthFilter(crop.Name, "Crop Borders", "Crop(%crop_left%,%crop_top%,-%crop_right%,-%crop_bottom%)", True))
        l.Add(crop)

        Dim field As New AviSynthCategory("Field")
        field.Filters.Add(New AviSynthFilter(field.Name, "FieldDeinterlace", "FieldDeinterlace()", True))
        field.Filters.Add(New AviSynthFilter(field.Name, "TomsMoComp", "TomsMoComp(-1,5,1)", True))
        field.Filters.Add(New AviSynthFilter(field.Name, "Yadif", "LoadCplugin(""%application:Yadif%"") Yadif()", True))
        field.Filters.Add(New AviSynthFilter(field.Name, "IVTC", "Telecide(guide=1).Decimate()", True))
        l.Add(field)

        Dim resize As New AviSynthCategory("Resize")
        resize.Filters.Add(New AviSynthFilter(resize.Name, "Soft", "BilinearResize(%target_width%,%target_height%)", True))
        resize.Filters.Add(New AviSynthFilter(resize.Name, "Neutral", "BicubicResize(%target_width%,%target_height%,0,0.5)", True))
        resize.Filters.Add(New AviSynthFilter(resize.Name, "Sharp", "LanczosResize(%target_width%,%target_height%)", True))
        resize.Filters.Add(New AviSynthFilter(resize.Name, "Very Sharp", "Lanczos4Resize(%target_width%,%target_height%)", True))
        l.Add(resize)

        Dim noise As New AviSynthCategory("Noise")
        noise.Filters.Add(New AviSynthFilter(noise.Name, "Low", "FluxSmoothST(2)", True))
        noise.Filters.Add(New AviSynthFilter(noise.Name, "Medium", "FluxSmoothST(4)", True))
        noise.Filters.Add(New AviSynthFilter(noise.Name, "Heavy", "FluxSmoothST(8)", True))
        noise.Filters.Add(New AviSynthFilter(noise.Name, "Extreme", "UnDot.Deen()", True))
        l.Add(noise)

        Dim subtitle As New AviSynthCategory("Subtitle")
        subtitle.Filters.Add(New AviSynthFilter(subtitle.Name, "VobSub", "VobSub()", True))
        subtitle.Filters.Add(New AviSynthFilter(subtitle.Name, "TextSub", "TextSub()", True))
        l.Add(subtitle)

        Dim audio As New AviSynthCategory("Audio")
        audio.Filters.Add(New AviSynthFilter(audio.Name, "AC3 Source", "AudioDub(last, NicAC3Source(""%audio_file1%""))", True))
        audio.Filters.Add(New AviSynthFilter(audio.Name, "MP123 Source", "AudioDub(last, NicMPG123Source(""%audio_file1%""))", True))
        l.Add(audio)

        Dim misc As New AviSynthCategory("Misc")
        misc.Filters.Add(New AviSynthFilter(misc.Name, "Convert To YV12", "ConvertToYV12()", True))

        l.Add(misc)

        Return l
    End Function
End Class

<Serializable()> _
Public Class AviSynthCategoryLegacy
    Public Sub New(ByVal name As String)
        Me.Name = name
    End Sub

    Private NameValue As String

    Public Property Name() As String
        Get
            Return NameValue
        End Get
        Set(ByVal Value As String)
            NameValue = Value
        End Set
    End Property

    Private FitersValue As New ArrayList

    Public Property Filters() As ArrayList
        Get
            Return FitersValue
        End Get
        Set(ByVal Value As ArrayList)
            FitersValue = Value
        End Set
    End Property

    Public Overrides Function ToString() As String
        Return Name
    End Function
End Class